@@ -1,9 +1,7 @@
 /*
     GQ -- a GTK-based LDAP client is
-    Copyright (C) 1998-2001 Bert Vermeulen
-    
-    This file (browse-dnd.c) is
-    Copyright (C) 2002 by Peter Stamfest <peter@stamfest.at>
+    Copyright (C) 1998-2003 Bert Vermeulen
+    Copyright (C) 2002-2003 Peter Stamfest
 
     This program is released under the Gnu General Public License with
     the additional exemption that compiling, linking, and/or using
@@ -24,9 +22,8 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-/* $Id: browse-dnd.c,v 1.6 2002/06/18 22:07:14 stamfest Exp $ */
+/* $Id: browse-dnd.c,v 1.25 2005/05/05 04:48:16 dave_malcolm Exp $ */
 
-/*  #define DEBUG_DND */
 /*  typedef unsigned long ulong; */
 
 #include "config.h"
@@ -38,6 +35,7 @@
 #include <gtk/gtk.h>
 
 #include "browse.h" 
+#include "dn-browse.h" 
 #include "browse-dnd.h" 
 
 #include "errorchain.h"
@@ -45,9 +43,10 @@
 #include "ldif.h"
 #include "util.h" 
 #include "i18n.h" 
+#include "debug.h" 
 
 typedef struct _dnd_refresh {
-     GtkCTree *tree;
+     GQTreeWidget *tree;
      struct ldapserver *server;
      int options;
      char *dn;
@@ -56,7 +55,7 @@ typedef struct _dnd_refresh {
 static GList *dnd_refresh_list = NULL;
 
 
-static dnd_refresh *new_dnd_refresh_dn(GtkCTree *tree, 
+static dnd_refresh *new_dnd_refresh_dn(GQTreeWidget *tree, 
 				       struct ldapserver *server,
 				       char *dn,
 				       int options)
@@ -75,12 +74,12 @@ static dnd_refresh *new_dnd_refresh_dn(GtkCTree *tree,
      return dr;
 }
 
-static dnd_refresh *new_dnd_refresh_node(GtkCTree *tree, 
-					 GtkCTreeNode *node,
+static dnd_refresh *new_dnd_refresh_node(GQTreeWidget *tree, 
+					 GQTreeWidgetNode *node,
 					 int options)
 {
      dn_browse_entry *entry;
-     entry = (dn_browse_entry *) gtk_ctree_node_get_row_data(tree, node);
+     entry = (dn_browse_entry *) gq_tree_get_node_data (tree, node);
 
      if (IS_DN_ENTRY(entry)) {
 	  return new_dnd_refresh_dn(tree, server_from_node(tree, node),
@@ -107,24 +106,21 @@ static gboolean drag_drop(GtkWidget *ctreeroot,
 			  guint time,
 			  gpointer user_data)
 {
-     int rc, row, column;
      GNode *node;
-     GtkCTreeNode *ctree_node;
+     GQTreeWidgetNode *ctree_node;
      struct node_entry *target_entry;
      int answer;
 
-#ifdef DEBUG_DND
-     printf("dropped %08lx at x=%d, y=%d\n", (ulong) drag_context, x, y);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSE_DND) {
+	  printf("dropped %08lx at x=%d, y=%d\n", (unsigned long) drag_context, x, y);
+     }
 #endif
 
-     rc = gtk_clist_get_selection_info(GTK_CLIST(ctreeroot), x, y,
-				       &row, &column);
-     if (rc == 0) return FALSE;
-     
-     ctree_node = gtk_ctree_node_nth(GTK_CTREE(ctreeroot), row);
+     ctree_node = gq_tree_get_node_at (GTK_CTREE(ctreeroot), x, y);
      if (ctree_node == NULL) return FALSE;
 
-     node = (GNode *) gtk_ctree_node_get_row_data(GTK_CTREE(ctreeroot),
+     node = (GNode *) gq_tree_get_node_data (GTK_CTREE(ctreeroot),
 						  GTK_CTREE_NODE(ctree_node));
      if (node == NULL) return FALSE;
      if ((target_entry = (struct node_entry *) node->data) == NULL) 
@@ -187,7 +183,7 @@ static int find_scroll_interval(int distance) {
      static int intervalmap[] = { 10, 10, 20, 20, 25, 25, 50, 50, 
 				  67, 100, 200 };
 
-     if (distance >= (sizeof(intervalmap) / sizeof(intervalmap[0])) ||
+     if (distance >= (int) (sizeof(intervalmap) / sizeof(intervalmap[0])) ||
 	 distance < 0) {
 	  return 200;
      } 
@@ -282,26 +278,28 @@ static void drag_and_select_data_free(GHashTable *hash)
 
 static void drag_begin(GtkWidget *ctreeroot,
 		       GdkDragContext *drag_context,
-		       GHashTable *hash) 
+		       struct tab *tab) 
 {
      /* store data we began dragging with for drag_data_get later on */
-     GtkCTreeNode *ctree_node;
+     GQTreeWidgetNode *ctree_node;
      struct ldapserver *server;
      browse_entry *entry;
      GHashTable *seldata;
      char cbuf[20];
 
-     ctree_node = g_hash_table_lookup(hash, "tree-row-selected");
-/*       node = (GNode *) gtk_ctree_node_get_row_data(GTK_CTREE(ctreeroot), */
+     ctree_node = BROWSETAB(tab)->tree_row_selected;
+/*       node = (GNode *) gq_tree_get_node_data (GTK_CTREE(ctreeroot), */
 /*  						  GTK_CTREE_NODE(ctree_node)); */
      server = server_from_node(GTK_CTREE(ctreeroot), ctree_node);
-     entry = (browse_entry *) gtk_ctree_node_get_row_data(GTK_CTREE(ctreeroot),
+     entry = (browse_entry *) gq_tree_get_node_data (GTK_CTREE(ctreeroot),
 							  ctree_node);
 
-#ifdef DEBUG_DND
-     printf("drag_begin start node=%08lx entry=%08lx\n",
-	    (ulong) ctree_node,
-	    (ulong) entry);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("drag_begin start node=%08lx entry=%08lx\n",
+		 (unsigned long) ctree_node,
+		 (unsigned long) entry);
+     }
 #endif
 
      if (IS_DN_ENTRY(entry)) {
@@ -324,13 +322,16 @@ static void drag_begin(GtkWidget *ctreeroot,
 				   seldata,
 				   (GtkDestroyNotify) drag_and_select_data_free);
 
-#ifdef DEBUG_DND
-	  printf("drag_begin %08lx %08lx\n", (ulong) ctreeroot,
-		 (ulong) seldata);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("drag_begin %08lx %08lx\n", (unsigned long) ctreeroot,
+		 (unsigned long) seldata);
+     }
 #endif
      }
 }
 
+
 static int dnd_refresh_sort_func(const dnd_refresh *a, const dnd_refresh *b)
 {
      if (a == NULL || a->dn == NULL) return -1;
@@ -339,7 +340,8 @@ static int dnd_refresh_sort_func(const dnd_refresh *a, const dnd_refresh *b)
      return strlen(a->dn) - strlen(b->dn);
 }
 
-static void do_refresh() {
+
+static void do_refresh(int error_context, struct tab *tab) {
      GList *l;
      dnd_refresh *at_end = NULL;
 
@@ -350,11 +352,14 @@ static void do_refresh() {
      dnd_refresh_list = g_list_sort(dnd_refresh_list,
 				    (GCompareFunc) dnd_refresh_sort_func);
 
+     set_update_lock(tab);
+
      for (l = g_list_first(dnd_refresh_list) ; l ; l = g_list_next(l)) {
 	  dnd_refresh *dr = l->data;
 	  if (dr->dn == NULL) continue;
 
-	  refresh_subtree_with_options(dr->tree, 
+	  refresh_subtree_with_options(error_context,
+				       dr->tree, 
 				       tree_node_from_server_dn(dr->tree, 
 								dr->server,
 								dr->dn),
@@ -365,8 +370,10 @@ static void do_refresh() {
 	  }
      }
 
+     release_update_lock(tab);
+
      if (at_end) {
-	  GtkCTreeNode *node = tree_node_from_server_dn(at_end->tree, 
+	  GQTreeWidgetNode *node = tree_node_from_server_dn(at_end->tree, 
 							at_end->server,
 							at_end->dn);
 /*  	  show_dn(at_end->tree, */
@@ -374,7 +381,7 @@ static void do_refresh() {
 
 /*  		  , at_end->dn); */
 	  if (node) 
-	       gtk_ctree_select(at_end->tree, node);
+	       gq_tree_select_node(at_end->tree, node);
      }
      
      for (l = g_list_first(dnd_refresh_list) ; l ; l = g_list_next(l)) {
@@ -387,10 +394,13 @@ static void do_refresh() {
 
 static void drag_end(GtkWidget *ctreeroot,
 		     GdkDragContext *drag_context,
-		     gpointer user_data) 
+		     struct tab *tab) 
 {
-#ifdef DEBUG_DND
-     printf("drag_end %08lx\n", (ulong) drag_context);
+     int ctx;
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("drag_end %08lx\n", (unsigned long) drag_context);
+     }
 #endif
      /* kill any motion timer ... */
      /* NOTE: This would not work for inter-widget drag/drop as the
@@ -403,7 +413,9 @@ static void drag_end(GtkWidget *ctreeroot,
      gtk_object_remove_data(GTK_OBJECT(ctreeroot), 
 			    "drag-and-selection-data");
 
-     do_refresh();
+     ctx = error_new_context(_("Refreshing entry after dragging"), ctreeroot);
+     do_refresh(ctx, tab);
+     error_flush(ctx);
 }
 
 
@@ -423,7 +435,7 @@ static void drag_data_get(GtkWidget *ctreeroot,
 			  GtkSelectionData *data,
 			  guint info,
 			  guint time,
-			  GHashTable *hash) 
+			  struct tab *tab)   /* don't really need tab here */
 {
      GHashTable *selhash = gtk_object_get_data(GTK_OBJECT(ctreeroot), 
 					       "drag-and-selection-data");
@@ -439,9 +451,11 @@ static void drag_data_get(GtkWidget *ctreeroot,
      g_hash_table_foreach(selhash, (GHFunc) drag_selection_data_prepare, buf);
      g_byte_array_append(buf, &nul, 1); 
 
-#ifdef DEBUG_DND
-     printf("drag_data_get ctx=%08lx data=%08lx\n", (ulong) drag_context,
-	    (ulong) data);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("drag_data_get ctx=%08lx data=%08lx\n", (unsigned long) drag_context,
+		 (unsigned long) data);
+     }
 #endif
 
      gtk_selection_data_set(data,
@@ -452,22 +466,20 @@ static void drag_data_get(GtkWidget *ctreeroot,
 
      g_byte_array_free(buf, TRUE);
 
-#ifdef DEBUG_DND
-     printf("data->data=%08lx data=%08lxx\n",
-	    (ulong)(data->data), (ulong) data);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("data->data=%08lx data=%08lxx\n",
+		 (unsigned long)(data->data), (unsigned long) data);
+     }
 #endif
 }
 
+
 static void move_progress(const char *from_dn,
 			  const char *to_dn,
 			  const char *new_dn) 
 {
-     int len = strlen(from_dn) + strlen(to_dn) + 80;
-     char *msg = g_malloc(len);
-/*       snprintf(msg, len, "Moved %s to %s", from_dn, to_dn); */
-     snprintf(msg, len, _("Created %s"), new_dn);
-     statusbar_msg(msg);
-     g_free(msg);
+     statusbar_msg(_("Created %s"), new_dn);
 }
 
 static GHashTable *drag_selection_data_unpack(const char *data, int len)
@@ -496,8 +508,10 @@ static GHashTable *drag_selection_data_unpack(const char *data, int len)
 	       strncpy(value, eq, n);
 	       value[n] = 0;
 
-#ifdef DEBUG_DND
-	       printf("key=%s value=%s\n", key, value);
+#ifdef DEBUG
+	       if (debug & GQ_DEBUG_BROWSER_DND) {
+		    printf("key=%s value=%s\n", key, value);
+	       }
 #endif
 	       g_hash_table_insert(selhash, key, value);
 	  } 
@@ -508,24 +522,30 @@ static GHashTable *drag_selection_data_unpack(const char *data, int len)
 
 static void do_move_after_reception(GtkWidget *ctreeroot,
 				    GHashTable *selhash,
-				    GtkCTreeNode *target_ctree_node,
+				    GQTreeWidgetNode *target_ctree_node,
 				    dn_browse_entry *target_entry,
 				    struct ldapserver *target_server,
 				    int flags)
 {
-     GtkCTreeNode *ctree_node;
+     GQTreeWidgetNode *ctree_node;
      char *dn = NULL;
      struct ldapserver *source_server;
      char *newdn = NULL;
+     int context;
 
-#ifdef DEBUG_DND
-     printf("do_move_after_reception selhash=%08lx server=%s dn=%s\n",
-	    (ulong) selhash, (char*)g_hash_table_lookup(selhash, "nickname"), dn);
+     context = error_new_context(_("Moving entry"), ctreeroot);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("do_move_after_reception selhash=%08lx server=%s dn=%s\n",
+		 (unsigned long) selhash,
+		 (const char*)g_hash_table_lookup(selhash, "nickname"), dn);
+     }
 #endif
 
      source_server = server_by_name(g_hash_table_lookup(selhash, "nickname"));
      if (source_server == NULL) {
-	  statusbar_msg(_("Could not find source server by its nickname!"));
+	  statusbar_msg(_("Could not find source server by its nickname ('%s')!"),
+			(const char*)g_hash_table_lookup(selhash, "nickname"));
 	  return;
      }
 
@@ -542,30 +562,25 @@ static void do_move_after_reception(GtkWidget *ctreeroot,
 
      if ((source_server == target_server) &&
 	 ( strcasecmp(dn, target_entry->dn) == 0)) {
-	  statusbar_msg(_("Cannot move/copy entry onto itself!"));
-          error_popup(_("Cannot move/copy entry onto itself!"), 
-		      _("Cannot move/copy entry onto itself!"));
-
-	  return;
+          error_push(context, _("Cannot move/copy entry onto itself!"));
+	  
+	  goto done;
      }
      
      if ((flags & MOVE_RECURSIVELY) &&
 	 (source_server == target_server) &&
 	 is_ancestor(target_entry->dn, dn)) {
-	  statusbar_msg(_("Cannot recursively move/copy entry onto or below itself!"));
-          error_popup(_("Cannot recursively move/copy entry onto or below itself!"), 
-		      _("Cannot recursively move/copy entry onto or below itself!"));
+          error_push(context,
+		     _("Cannot recursively move/copy entry onto or below itself!"));
 	  
-	  return;
+	  goto done;
      }
      
-
-
      if (IS_DN_ENTRY(target_entry)) {
 	  newdn = move_entry(dn, source_server, 
 			     target_entry->dn, target_server,
-			     flags, move_progress);
-	  
+			     flags, move_progress, context);
+
 	  /* register that we have to refresh the target node */
 	  dnd_refresh_list =
 	       g_list_append(dnd_refresh_list,
@@ -592,19 +607,23 @@ static void do_move_after_reception(GtkWidget *ctreeroot,
 	       ctree_node = node_from_dn(GTK_CTREE(ctreeroot),
 					 target_ctree_node, newdn);
 	       if (ctree_node)
-		    gtk_ctree_select(GTK_CTREE(ctreeroot),
-				     ctree_node);
+		    gq_tree_select_node(GQ_TREE_WIDGET(ctreeroot),
+					ctree_node);
 	       
 
-#ifdef DEBUG_DND
-	       printf("moved %s below %s, became %s\n",
-		      dn,
-		      target_entry->dn,
-		      newdn);
+#ifdef DEBUG
+	       if (debug & GQ_DEBUG_BROWSER_DND) {
+		    printf("moved %s below %s, became %s\n",
+			   dn,
+			   target_entry->dn,
+			   newdn);
+	       }
 #endif
 	       g_free(newdn);
 	  }
      }
+ done:
+     error_flush(context);
 }
 
 /* handles data receiving and does the actual move */
@@ -617,9 +636,8 @@ static void drag_data_received(GtkWidget *ctreeroot,
 			       guint time,
 			       gpointer user_data)
 {
-     GtkCTreeNode *ctree_node;
+     GQTreeWidgetNode *ctree_node;
      struct ldapserver *target_server;
-     int rc, row, column;
      GHashTable *selhash = NULL;
      dn_browse_entry *target_entry;
      int answer;
@@ -629,21 +647,19 @@ static void drag_data_received(GtkWidget *ctreeroot,
 	  return;
      }
 
-#ifdef DEBUG_DND
-     printf("drag_data_received ctx=%08lx seldata=%08lx\n",
-	    (ulong) drag_context, (ulong) data);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("drag_data_received ctx=%08lx seldata=%08lx\n",
+		 (unsigned long) drag_context, (unsigned long) data);
+     }
 #endif
 
-     rc = gtk_clist_get_selection_info(GTK_CLIST(ctreeroot), x, y,
-				       &row, &column);
-     if (rc == 0) return;
-     
-     ctree_node = gtk_ctree_node_nth(GTK_CTREE(ctreeroot), row);
+     ctree_node = gq_tree_get_node_at (GTK_CTREE(ctreeroot), x, y);
      if (ctree_node == NULL) return;
 
      /* NOTE: type has not yet been checked !!! */
      target_entry = (dn_browse_entry *) 
-	  gtk_ctree_node_get_row_data(GTK_CTREE(ctreeroot),
+	  gq_tree_get_node_data (GTK_CTREE(ctreeroot),
 				      GTK_CTREE_NODE(ctree_node));
 
      if (! IS_DN_ENTRY(target_entry)) return;
@@ -678,8 +694,10 @@ static void drag_data_received(GtkWidget *ctreeroot,
 
 	  }
      }
-#ifdef DEBUG_DND
-     printf("dragged to %d/%d\n", x, y);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("dragged to %d/%d\n", x, y);
+     }
 #endif
 }
 
@@ -697,31 +715,39 @@ static void drag_data_received(GtkWidget *ctreeroot,
 
 static void drag_data_delete(GtkWidget *widget,
 			     GdkDragContext *drag_context,
-			     GHashTable *hash)
+			     struct tab *tab)
 {
-     GtkCTreeNode *ctree_node;
-     char *server = NULL, *dn = NULL;
+     GQTreeWidgetNode *ctree_node;
+     char *s = NULL, *dn = NULL;
+     struct ldapserver *server;
 
      GHashTable *selhash = gtk_object_get_data(GTK_OBJECT(widget), 
 					       "drag-and-selection-data");
 
-     server = g_hash_table_lookup(selhash, "nickname");
+     s = g_hash_table_lookup(selhash, "nickname");
      dn = g_hash_table_lookup(selhash, "dn");
 
-#ifdef DEBUG_DND
-     printf("drag_data_delete ctx=%08lx suggested=%d action=%d server=%s dn=%s\n", 
-	    (ulong) drag_context,
-	    drag_context->suggested_action,
-	    drag_context->action,
-	    server, dn
-	    );
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("drag_data_delete ctx=%08lx suggested=%d action=%d server=%s dn=%s\n", 
+		 (unsigned long) drag_context,
+		 drag_context->suggested_action,
+		 drag_context->action,
+		 s, dn
+		 );
+     }
 #endif
 
-     ctree_node = tree_node_from_server_dn(GTK_CTREE(widget),
-					   server_by_name(server), dn);
+     server = server_by_name(s);
+     if (!server) {
+	  return;
+     }
+     ctree_node = tree_node_from_server_dn(GQ_TREE_WIDGET(widget),
+					   server, dn);
      
      if (ctree_node) {
-	  GtkCTreeNode *parent = GTK_CTREE_ROW(ctree_node)->parent;
+	  GQTreeWidgetNode *parent = gq_tree_get_parent_node (GQ_TREE_WIDGET (widget),
+							      ctree_node);
   	  if (parent) {
 	       /* register that we have to refresh the parent node */
 
@@ -733,18 +759,21 @@ static void drag_data_delete(GtkWidget *widget,
 /*    	       refresh_subtree_with_options(GTK_CTREE(widget), parent, 0); */
 	  }
      }
-#ifdef DEBUG_DND
-     printf("drag_data_delete done\n");
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("drag_data_delete done\n");
+     }
 #endif
      gtk_object_remove_data(GTK_OBJECT(widget), 
 			    "drag-and-selection-data");
 }
 
+
 #if 0
-gboolean compare_drag_func(GtkCTree *ctreeroot,
-			   GtkCTreeNode *source_node,
-			   GtkCTreeNode *new_parent,
-			   GtkCTreeNode *new_sibling)
+gboolean compare_drag_func(GQTreeWidget *ctreeroot,
+			   GQTreeWidgetNode *source_node,
+			   GQTreeWidgetNode *new_parent,
+			   GQTreeWidgetNode *new_sibling)
 {
     printf("compare_drag_func\n");
     return TRUE;
@@ -752,21 +781,20 @@ gboolean compare_drag_func(GtkCTree *ctreeroot,
 #endif
 
 
-void copy_entry(GtkWidget *widget, GHashTable *hash)
+void copy_entry(GtkWidget *widget, struct tab *tab)
 {
-     GtkCTree *ctree;
+     GQTreeWidget *ctree;
      int have_sel;
      struct ldapserver *server;
      dn_browse_entry *entry;
-     GtkCTreeNode *ctree_node;
+     GQTreeWidgetNode *ctree_node;
      GHashTable *seldata;
      char cbuf[20];
 
-     ctree = g_hash_table_lookup(hash, "ctreeroot");
+     ctree = BROWSETAB(tab)->ctreeroot;
 
-     ctree_node = g_hash_table_lookup(hash, "tree-row-popped-up");
-     entry = (dn_browse_entry *) gtk_ctree_node_get_row_data(ctree,
-							     ctree_node);
+     ctree_node = BROWSETAB(tab)->tree_row_popped_up;
+     entry = (dn_browse_entry *) gq_tree_get_node_data (ctree, ctree_node);
      if (entry == NULL) return;
      if (!IS_DN_ENTRY(entry)) return;
 
@@ -774,10 +802,8 @@ void copy_entry(GtkWidget *widget, GHashTable *hash)
 
      seldata = g_hash_table_new(g_str_hash, g_str_equal);
      g_hash_table_insert(seldata, g_strdup("dn"), g_strdup(entry->dn));
-     g_hash_table_insert(seldata, g_strdup("nickname"), 
-			 g_strdup(server->name));
-     g_hash_table_insert(seldata, g_strdup("server"), 
-			 g_strdup(server->ldaphost));
+     g_hash_table_insert(seldata, g_strdup("nickname"), g_strdup(server->name));
+     g_hash_table_insert(seldata, g_strdup("server"), g_strdup(server->ldaphost));
      snprintf(cbuf, sizeof(cbuf), "%d", server->ldapport);
      g_hash_table_insert(seldata, g_strdup("port"), g_strdup(cbuf));
 
@@ -790,30 +816,38 @@ void copy_entry(GtkWidget *widget, GHashTable *hash)
 					GDK_SELECTION_PRIMARY,
 					GDK_CURRENT_TIME);
 
-#ifdef DEBUG_DND
-     printf("copy_entry %08lx %d\n", (ulong) ctree, have_sel);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("copy_entry %08lx %d\n", (unsigned long) ctree, have_sel);
+     }
 #endif
+
 }
 
-void copy_entry_all(GtkWidget *widget, GHashTable *hash)
+
+void copy_entry_all(GtkWidget *widget, struct tab *tab)
 {
      GHashTable *selhash = NULL;
      GtkWidget *ctreeroot;
 
-     copy_entry(widget, hash);
+     copy_entry(widget, tab);
 
-     ctreeroot = g_hash_table_lookup(hash, "ctreeroot");
+     ctreeroot = GTK_WIDGET(BROWSETAB(tab)->ctreeroot);
 
      selhash = gtk_object_get_data(GTK_OBJECT(ctreeroot), 
 				   "drag-and-selection-data");
      
      g_hash_table_insert(selhash, g_strdup("recursively"), g_strdup("TRUE"));
 
-#ifdef DEBUG_DND
-     printf("copy_entry_all %08lx\n", (ulong) widget);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("copy_entry_all %08lx\n", (unsigned long) widget);
+     }
 #endif
+
 }
 
+
 static void get_selection_string(GtkWidget *ctree,
 				 GtkSelectionData *data,
 				 GHashTable *selhash)
@@ -821,49 +855,94 @@ static void get_selection_string(GtkWidget *ctree,
      struct ldapserver *server = 
 	  server_by_name(g_hash_table_lookup(selhash, "nickname"));
      char *dn = g_hash_table_lookup(selhash, "dn");
-
+     LDAPControl c, *ctrls[2] = { NULL, NULL } ;
+     int scope;
      LDAP *ld;
      int rc;
 
-     LDAPMessage *msg = NULL, *e;
+     LDAPMessage *e, *res = NULL;
      GString *out = NULL;
      gboolean ok = FALSE;
 
-#ifdef DEBUG_DND
-     printf("get_selection_string\n");
+     char *attrs[] = {
+	  LDAP_ALL_USER_ATTRIBUTES,
+	  "ref",
+	  NULL 
+     };
+	       
+     int ctx = error_new_context(_("Getting selection string"), ctree);
+
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("get_selection_string\n");
+     }
 #endif
 
+     if (server == NULL) {
+	  error_push(ctx,
+		     _("Cannot find server by its nickname ('%s')"), 
+		     (char*) g_hash_table_lookup(selhash, "nickname"));
+	  goto done;
+     }
      /* retrieve from server ... */
 
-     if( (ld = open_connection(server)) == NULL) goto fail;
+     if( (ld = open_connection(ctx, server)) == NULL) goto fail;
 
-     rc = ldap_search_s(ld, dn, 
-			g_hash_table_lookup(selhash, "recursively") ? LDAP_SCOPE_SUBTREE : LDAP_SCOPE_BASE,
-			"objectClass=*", NULL, 0, &msg);
-
-     if (rc == LDAP_SERVER_DOWN) {
-	  server->server_down = 1;
-	  goto fail;
-     }
-     if (rc != LDAP_SUCCESS) goto fail;
-
-     out = g_string_new("");
-
-     for(e = ldap_first_entry(ld, msg); e ; e = ldap_next_entry(ld, e)) {
-	  ldif_entry_out(out, ld, e);
-     }
+     /* use ManageDSAit: we want to move/copy refs intact */
+     c.ldctl_oid		= LDAP_CONTROL_MANAGEDSAIT;
+     c.ldctl_value.bv_val	= NULL;
+     c.ldctl_value.bv_len	= 0;
+     c.ldctl_iscritical	= 1;
      
-     gtk_selection_data_set(data,
-			    GDK_SELECTION_TYPE_STRING,
-			    8, /* character data */
-			    out->str,
-			    out->len
-			    );
-     ok = TRUE;
+     ctrls[0] = &c;
+
+     scope = g_hash_table_lookup(selhash, "recursively") ? 
+	  LDAP_SCOPE_SUBTREE : LDAP_SCOPE_BASE;
+
+     rc = ldap_search_ext_s(ld, 
+			    dn,			/* base */
+			    scope,		/* search scope */
+			    "(objectClass=*)",	/* filter */
+			    attrs,		/* attrs */
+			    0,			/* attrsonly */
+			    ctrls,		/* serverctrls */
+			    NULL,		/* clientctrls */
+			    NULL,		/* timeout */
+			    LDAP_NO_LIMIT,	/* sizelimit */
+			    &res);
+
+     if(rc == LDAP_NOT_SUPPORTED) {
+	  rc = ldap_search_s(ld, dn, scope, "(objectClass=*)", attrs, 0, &res);
+     }
+
+     if (rc == LDAP_SUCCESS) { 
+	  out = g_string_new("");
+
+	  for( e = ldap_first_entry(ld, res) ; e != NULL ;
+	       e = ldap_next_entry(ld, e) ) {
+	       ldif_entry_out(out, ld, e, ctx);
+	  }
+	  
+	  gtk_selection_data_set(data,
+				 GDK_SELECTION_TYPE_STRING,
+				 8, /* character data */
+				 out->str,
+				 out->len
+				 );
+	  ok = TRUE;
+     } else if (rc == LDAP_SERVER_DOWN) {
+	  server->server_down++;
+	  error_push(ctx, 
+		     _("Error searching below '%1$s': %2$s"),
+		     dn, ldap_err2string(rc));
+	  push_ldap_addl_error(ld, ctx); 
+     } else {
+	  error_push(ctx, 
+		     _("Error searching below '%1$s': %2$s"),
+		     dn, ldap_err2string(rc));
+	  push_ldap_addl_error(ld, ctx); 
+     }
 
-     /* I hate to do it like this, but sometimes labels are such
-        helpful beasts */
- fail:
      if (!ok) {
 	  gtk_selection_data_set(data,
 				 GDK_SELECTION_TYPE_STRING,
@@ -873,12 +952,17 @@ static void get_selection_string(GtkWidget *ctree,
 				 );
      }
 
+ fail:
      if (out) g_string_free(out, TRUE);
-     if (msg) ldap_msgfree(msg);
-     close_connection(server, FALSE);
+     if (res) ldap_msgfree(res); 
+     if (ld) close_connection(server, FALSE);
+
+ done:
+     error_flush(ctx);
      return;
 }
 
+
 static void get_selection_gq(GtkWidget *ctree,
 			     GtkSelectionData *data,
 			     GHashTable *selhash)
@@ -889,8 +973,10 @@ static void get_selection_gq(GtkWidget *ctree,
      g_hash_table_foreach(selhash, (GHFunc) drag_selection_data_prepare, buf);
      g_byte_array_append(buf, &nul, 1); 
 
-#ifdef DEBUG_DND
-     printf("get_selection_gq data=%08lx\n", (ulong) data);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("get_selection_gq data=%08lx\n", (unsigned long) data);
+     }
 #endif
 
      gtk_selection_data_set(data,
@@ -899,19 +985,22 @@ static void get_selection_gq(GtkWidget *ctree,
 			    buf->data,
 			    buf->len);
 
-#ifdef DEBUG_DND
-     printf("data->data=%08lx data=%08lxx\n",
-	    (ulong)(data->data), (ulong) data);
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("data->data=%08lx data=%08lxx\n",
+		 (unsigned long)(data->data), (unsigned long) data);
+     }
 #endif
 
      g_byte_array_free(buf, TRUE);
 }
 
+
 static void get_selection(GtkWidget *ctree,
 			  GtkSelectionData *data,
 			  guint info,
 			  guint time,
-			  GHashTable *hash)
+			  struct tab *tab)  /* XXX don't need to pass tab here */
 {
      GHashTable *selhash = gtk_object_get_data(GTK_OBJECT(ctree), 
 					       "drag-and-selection-data");
@@ -923,20 +1012,31 @@ static void get_selection(GtkWidget *ctree,
 	       get_selection_string(ctree, data, selhash);
 	  }
      }
+
 }
 
-void paste_entry(GtkWidget *widget, GHashTable *hash)
+
+void paste_entry(GtkWidget *widget, struct tab *tab)
 {
-     GtkWidget *ctree = g_hash_table_lookup(hash, "ctreeroot");
-     
-#ifdef DEBUG_DND
-     printf("paste_entry %08lx\n", (ulong) ctree);
+     GtkWidget *ctree;
+     int ctx;
+
+     ctree = GTK_WIDGET(BROWSETAB(tab)->ctreeroot);
+
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("paste_entry %08lx\n", (unsigned long) ctree);
+     }
 #endif
      gtk_selection_convert(ctree,
 			   GDK_SELECTION_PRIMARY,
 			   gdk_atom_intern("gq-browse-ctree", FALSE),
 			   GDK_CURRENT_TIME);
-     do_refresh();
+
+
+     ctx = error_new_context(_("Refreshing entry after pasting"), ctree);
+     do_refresh(ctx, tab);
+     error_flush(ctx);
 }
 
 /* gets fired after the gtk_selection_convert triggered through
@@ -944,24 +1044,27 @@ void paste_entry(GtkWidget *widget, GHashTable *hash)
 static void selection_received(GtkWidget *ctree,
 			       GtkSelectionData *data,
 			       guint time,
-			       GHashTable *hash) 
+			       struct tab *tab) 
 {
-
-     GtkCTreeNode *ctree_node;
+     GQTreeWidgetNode *ctree_node;
      dn_browse_entry *target_entry;
      struct ldapserver *target_server;
-     GtkCTree *ctreeroot = g_hash_table_lookup(hash, "ctreeroot");
+     GQTreeWidget *ctreeroot;
      GHashTable *selhash = NULL;
 
-#ifdef DEBUG_DND
-     printf("selection_received seldata=%08lx\n",
-	    (ulong) data);
+     ctreeroot = BROWSETAB(tab)->ctreeroot;
+
+#ifdef DEBUG
+     if (debug & GQ_DEBUG_BROWSER_DND) {
+	  printf("selection_received seldata=%08lx\n",
+		 (unsigned long) data);
+     }
 #endif
 
-     ctree_node = g_hash_table_lookup(hash, "tree-row-popped-up");
+     ctree_node = BROWSETAB(tab)->tree_row_popped_up;
      target_entry = 
-	  (dn_browse_entry *) gtk_ctree_node_get_row_data(ctreeroot,
-							  ctree_node);
+	  (dn_browse_entry *) gq_tree_get_node_data (ctreeroot,
+						     ctree_node);
      if (target_entry == NULL) return;
      if (!IS_DN_ENTRY(target_entry)) return;
      
@@ -1010,7 +1113,8 @@ static void selection_received(GtkWidget *ctree,
      }
 }
 
-void browse_dnd_setup(GtkWidget *ctreeroot, GHashTable *hash) 
+
+void browse_dnd_setup(GtkWidget *ctreeroot, struct tab *tab) 
 {
      GtkTargetEntry *target_entry;
 
@@ -1027,11 +1131,11 @@ void browse_dnd_setup(GtkWidget *ctreeroot, GHashTable *hash)
      gtk_signal_connect(GTK_OBJECT(ctreeroot), 
 			"selection-get",
 			(GtkSignalFunc) get_selection,
-			(gpointer) hash);
+			(gpointer) tab);
      gtk_signal_connect(GTK_OBJECT(ctreeroot), 
 			"selection-received",
 			(GtkSignalFunc) selection_received,
-			(gpointer) hash);
+			(gpointer) tab);
 
      target_entry = (GtkTargetEntry*) g_malloc(sizeof(GtkTargetEntry));
      target_entry->target = "gq-browse-ctree";
@@ -1057,11 +1161,11 @@ void browse_dnd_setup(GtkWidget *ctreeroot, GHashTable *hash)
 
      gtk_signal_connect(GTK_OBJECT(ctreeroot), "drag-data-get",
 			 GTK_SIGNAL_FUNC(drag_data_get),
-			 (gpointer) hash);
+			(gpointer) tab);
      
      gtk_signal_connect(GTK_OBJECT(ctreeroot), "drag-data-delete",
 			 GTK_SIGNAL_FUNC(drag_data_delete),
-			 (gpointer) hash);
+			 (gpointer) tab);
 
      gtk_signal_connect(GTK_OBJECT(ctreeroot), "drag-motion",
 			GTK_SIGNAL_FUNC(drag_motion),
@@ -1069,13 +1173,14 @@ void browse_dnd_setup(GtkWidget *ctreeroot, GHashTable *hash)
 
      gtk_signal_connect(GTK_OBJECT(ctreeroot), "drag-begin",
 			GTK_SIGNAL_FUNC(drag_begin),
-			(gpointer) hash);
+			(gpointer) tab);
 
 
      /* HACK alert */
      gtk_signal_connect(GTK_OBJECT(ctreeroot), "drag-end",
 			GTK_SIGNAL_FUNC(drag_end),
-			(gpointer) NULL);
+			(gpointer) tab);
+
 #if 0
      gtk_signal_connect(GTK_OBJECT(ctreeroot), "drag-drop",
 			GTK_SIGNAL_FUNC(drag_drop),
@@ -1086,6 +1191,7 @@ void browse_dnd_setup(GtkWidget *ctreeroot, GHashTable *hash)
 
      g_free(target_entry);
      target_entry = NULL;
+
 }
 
 #endif /* BROWSER_DND */
