@@ -1,6 +1,7 @@
 /*
     GQ -- a GTK-based LDAP client
-    Copyright (C) 1998-2001 Bert Vermeulen
+    Copyright (C) 1998-2003 Bert Vermeulen
+    Copyright (C) 2002-2003 Peter Stamfest
 
     This program is released under the Gnu General Public License with
     the additional exemption that compiling, linking, and/or using
@@ -24,11 +25,18 @@
 #ifndef GQ_LDIF_H_INCLUDED
 #define GQ_LDIF_H_INCLUDED
 
-void prepend_ldif_header(GString *out, struct ldapserver *server,
-			 GList *bases);
-gboolean ldif_entry_out(GString *out, LDAP *ld, LDAPMessage *msg);
-gboolean ldif_line_out(GString *out, char *attr, char *value, int vlen);
-void b64_encode(GString *out, char *value, int vlen);
-void b64_decode(GByteArray *out, const char *value, int vlen);
+#include <glib.h>
+#include <gtk/gtk.h>
+
+#include "common.h"
+
+void prepend_ldif_header(GString *out, GList *to_export);
+gboolean ldif_entry_out(GString *out, LDAP *ld, LDAPMessage *msg,
+			int error_context);
+gboolean ldif_line_out(GString *out, char *attr, char *value, 
+		       unsigned int vlen,
+		       int error_context);
+void b64_encode(GString *out, char *value, unsigned int vlen);
+void b64_decode(GByteArray *out, const char *value, unsigned int vlen);
 
 #endif
