@@ -1,6 +1,7 @@
 /*
     GQ -- a GTK-based LDAP client
-    Copyright (C) 1998-2001 Bert Vermeulen
+    Copyright (C) 1998-2003 Bert Vermeulen
+    Copyright (C) 2002-2003 Peter Stamfest
 
     This program is released under the Gnu General Public License with
     the additional exemption that compiling, linking, and/or using
@@ -21,6 +22,7 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+/* $Id: schemabrowse.c,v 1.31 2006/04/06 02:15:46 herzi Exp $ */
 
 #include <config.h>
 
@@ -32,29 +34,77 @@
 
 #include <string.h>
 
+
+#define GTK_ENABLE_BROKEN  /* for the tree widget - should be replaced */
+
 #include <glib.h>
 #include <gtk/gtk.h>
 
+#include "schemabrowse.h"
 #include "mainwin.h"
 #include "configfile.h"
 #include "common.h"
 #include "util.h"
-#include "schemabrowse.h"
 #include "schema.h"
 #include "i18n.h"
 #include "debug.h"
+#include "errorchain.h"
 
 
-extern GtkWidget *statusbar, *mainwin;
-extern struct gq_config config;
+static void attach_schema_item(GtkWidget *tree, const char *type,
+			       void *schemaobject, char *itemname);
+static gboolean schema_button_tree(GtkWidget *tree_item, GdkEventButton *event,
+				   GtkWidget *root);
 
 
+static void add_single_schema_server(struct tab *tab,
+				     struct ldapserver *server);
+static void add_schema_servers(struct tab *tab);
+static void attach_server_schema(GtkWidget *item, struct tab *tab);
+static void schema_refresh_server(GtkWidget *widget, GtkWidget *item);
 
-void new_schemamode(GHashTable *hash)
+static void make_detail_notebook(struct tab *tab);
+static void popup_detail_callback(GtkWidget *dummy, GtkWidget *widget);
+static void make_oc_detail(GtkWidget *target_oc_vbox);
+static void fill_oc_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item,
+				     struct tab *tab);
+static void fill_oc_detail(GtkWidget *target_oc_vbox,
+			   struct ldapserver *server,
+			   LDAPObjectClass *oc);
+
+static void make_at_detail(GtkWidget *target_at_vbox);
+static void fill_at_detail_rightpane(GtkWidget *treeroot,
+				     GtkWidget *tree_item, struct tab *tab);
+static void fill_at_detail(int error_context, GtkWidget *target_vbox,
+			   struct ldapserver *server,
+			   LDAPAttributeType *at);
+
+static void make_mr_detail(GtkWidget *target_mr_vbox);
+static void fill_mr_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item,
+				     struct tab *tab);
+static void fill_mr_detail(GtkWidget *target_vbox, struct ldapserver *server,
+			   LDAPMatchingRule *mr);
+
+static void make_s_detail(GtkWidget *target_vbox);
+static void fill_s_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item,
+				    struct tab *tab);
+static void fill_s_detail(GtkWidget *target_vbox, struct ldapserver *server,
+			  LDAPSyntax *s);
+
+static void destroy_schema_mode(GtkWidget *w, struct tab *tab);
+
+struct tab *new_schemamode()
 {
      GtkWidget *schemamode_vbox, *rightpane_vbox, *spacer;
      GtkWidget *mainpane, *treeroot;
      GtkWidget *leftpane_scrwin, *rightpane_scrwin;
+     struct tab_schema *modeinfo;
+
+     struct tab *tab = g_malloc0(sizeof(struct tab));
+     tab->type = SCHEMA_MODE;
+
+     modeinfo = calloc(sizeof(struct tab_schema), 1);
+     tab->modeinfo = modeinfo;
 
      schemamode_vbox = gtk_vbox_new(FALSE, 0);
 
@@ -68,8 +118,9 @@ void new_schemamode(GHashTable *hash)
      gtk_box_pack_start(GTK_BOX(schemamode_vbox), mainpane, TRUE, TRUE, 0);
 
      treeroot = gtk_tree_new();
+     modeinfo->treeroot = treeroot;
      gtk_widget_show(treeroot);
-     add_schema_servers(hash, treeroot);
+     add_schema_servers(tab);
 
      leftpane_scrwin = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(leftpane_scrwin),
@@ -90,7 +141,7 @@ void new_schemamode(GHashTable *hash)
 
      rightpane_vbox = gtk_vbox_new(FALSE, 5);
      gtk_widget_show(rightpane_vbox);
-     g_hash_table_insert(hash, "rightpane_vbox", rightpane_vbox);
+     modeinfo->rightpane_vbox = rightpane_vbox;
      gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(rightpane_scrwin),
                                            rightpane_vbox);
      gtk_paned_add2(GTK_PANED(mainpane), rightpane_scrwin);
@@ -98,46 +149,68 @@ void new_schemamode(GHashTable *hash)
      gtk_object_set_data(GTK_OBJECT(schemamode_vbox), "focus", rightpane_scrwin);
      gtk_widget_show(schemamode_vbox);
 
-     g_hash_table_insert(hash, "vbox", schemamode_vbox);
+
+     gtk_signal_connect(GTK_OBJECT(schemamode_vbox), "destroy",
+			GTK_SIGNAL_FUNC(destroy_schema_mode),
+			tab);
+
+     tab->content = schemamode_vbox;
+     gtk_object_set_data(GTK_OBJECT(tab->content), "tab", tab);
+     return tab;
 }
 
 
-void add_schema_servers(GHashTable *hash, GtkWidget *treeroot)
+static void add_schema_servers(struct tab *tab)
+{
+     struct ldapserver *server;
+     int server_cnt = 0;
+     GList *I;
+     
+     for (I = config->servers ; I ; I = g_list_next(I)) {
+	  server = (struct ldapserver *) I->data;
+	  add_single_schema_server(tab, server);
+	  server_cnt++;
+     }
+
+     statusbar_msg(ngettext("One server found", "%d servers found",
+			    server_cnt),
+		   server_cnt);
+}
+
+
+static void add_single_schema_server(struct tab *tab,
+				     struct ldapserver *server)
 {
      GtkWidget *new_item, *new_subtree;
-     struct ldapserver *server;
-     int server_cnt;
-     char message[128];
 
-     server_cnt = 0;
-     server = config.ldapservers;
-     while(server) {
+     if (server) {
 	  new_item = gtk_tree_item_new_with_label(server->name);
 	  GTK_WIDGET_UNSET_FLAGS(new_item, GTK_CAN_FOCUS);
 	  gtk_signal_connect(GTK_OBJECT(new_item), "button_press_event",
-			     GTK_SIGNAL_FUNC(schema_button_tree), NULL);
-	  gtk_object_set_data(GTK_OBJECT(new_item), "server", (gpointer) server);
+			     GTK_SIGNAL_FUNC(schema_button_tree), 
+			     SCHEMATAB(tab)->treeroot);
+	  gtk_signal_connect(GTK_OBJECT(new_item), "button_release_event",
+			     GTK_SIGNAL_FUNC(schema_button_tree), 
+			     SCHEMATAB(tab)->treeroot);
+	  gtk_object_set_data(GTK_OBJECT(new_item), "tab", tab);
+	  gtk_object_set_data_full(GTK_OBJECT(new_item), "server",
+				   (gpointer) server,
+				   (GtkDestroyNotify) ldapserver_unref);
+	  ldapserver_ref(server);
 
-	  gtk_tree_append(GTK_TREE(treeroot), new_item);
+	  gtk_tree_append(GTK_TREE(SCHEMATAB(tab)->treeroot), new_item);
 	  new_subtree = gtk_tree_new();
 	  gtk_widget_show(new_subtree);
 	  gtk_tree_item_set_subtree(GTK_TREE_ITEM(new_item), new_subtree);
 	  gtk_widget_show(new_item);
 	  gtk_signal_connect(GTK_OBJECT(new_item), "expand",
-			     GTK_SIGNAL_FUNC(attach_server_schema), hash);
+			     GTK_SIGNAL_FUNC(attach_server_schema), tab);
 
-	  server = server->next;
-	  server_cnt++;
      }
-
-     make_message(message, sizeof(message), 
-		  server_cnt, _("server"), _("servers"), _("found"));
-     statusbar_msg(message);
-
 }
 
 
-void attach_server_schema(GtkWidget *item, GHashTable *hash)
+static void attach_server_schema(GtkWidget *item, struct tab *tab)
 {
      LDAPObjectClass *oc;
      LDAPAttributeType *at;
@@ -148,16 +221,18 @@ void attach_server_schema(GtkWidget *item, GHashTable *hash)
      struct ldapserver *server;
      struct server_schema *ss;
      int i;
+     int attach_context = error_new_context(_("Expanding server schema entry"),
+					    item);
 
      gtk_signal_disconnect_by_func(GTK_OBJECT(item),
-				   GTK_SIGNAL_FUNC(attach_server_schema), hash);
+				   GTK_SIGNAL_FUNC(attach_server_schema), tab);
 
      if( (server = gtk_object_get_data(GTK_OBJECT(item), "server")) == NULL)
 	  return;
 
      set_busycursor();
 
-     ss = get_server_schema(server);
+     ss = get_server_schema(attach_context, server);
      if(ss) {
 
 	  server_subtree = GTK_TREE_ITEM_SUBTREE(GTK_TREE_ITEM(item));
@@ -165,26 +240,36 @@ void attach_server_schema(GtkWidget *item, GHashTable *hash)
 	       new_item = gtk_tree_item_new_with_label("objectClasses");
 	       GTK_WIDGET_UNSET_FLAGS(new_item, GTK_CAN_FOCUS);
 	       gtk_signal_connect(GTK_OBJECT(new_item), "button_press_event",
-				  GTK_SIGNAL_FUNC(schema_button_tree), NULL);
+				  GTK_SIGNAL_FUNC(schema_button_tree), 
+				  SCHEMATAB(tab)->treeroot);
+	       gtk_signal_connect(GTK_OBJECT(new_item), "button_release_event",
+				  GTK_SIGNAL_FUNC(schema_button_tree), 
+				  SCHEMATAB(tab)->treeroot);
 	       gtk_widget_show(new_item);
 	       gtk_tree_append(GTK_TREE(server_subtree), new_item);
 
 	       new_subtree = gtk_tree_new();
 	       GTK_WIDGET_UNSET_FLAGS(new_subtree, GTK_CAN_FOCUS);
-	       gtk_object_set_data(GTK_OBJECT(new_subtree), "server", server);
+	       gtk_object_set_data_full(GTK_OBJECT(new_subtree), "server",
+					server,
+					(GtkDestroyNotify) ldapserver_unref);
+	       ldapserver_ref(server);
+
 	       gtk_widget_show(new_subtree);
 	       gtk_tree_item_set_subtree(GTK_TREE_ITEM(new_item), new_subtree);
-	       gtk_signal_connect(GTK_OBJECT(new_subtree), "select_child",
-				  GTK_SIGNAL_FUNC(fill_oc_detail_rightpane), hash);
+	       gtk_signal_connect(GTK_OBJECT(new_subtree), "select-child",
+				  GTK_SIGNAL_FUNC(fill_oc_detail_rightpane), tab);
 
 	       tmp = ss->oc;
 	       while(tmp) {
 		    oc = (LDAPObjectClass *) tmp->data;
-		    if(oc->oc_names && oc->oc_names[0])
-			 for(i = 0; oc->oc_names[i]; i++)
+		    if(oc->oc_names && oc->oc_names[0]) {
+			 for(i = 0; oc->oc_names[i]; i++) {
 			      attach_schema_item(new_subtree, "oc", oc, oc->oc_names[i]);
-		    else
+			 }
+		    } else {
 			 attach_schema_item(new_subtree, "oc", oc, oc->oc_oid);
+		    }
 		    tmp = tmp->next;
 	       }
 	  }
@@ -193,26 +278,36 @@ void attach_server_schema(GtkWidget *item, GHashTable *hash)
 	       new_item = gtk_tree_item_new_with_label("attributeTypes");
 	       GTK_WIDGET_UNSET_FLAGS(new_item, GTK_CAN_FOCUS);
 	       gtk_signal_connect(GTK_OBJECT(new_item), "button_press_event",
-				  GTK_SIGNAL_FUNC(schema_button_tree), NULL);
+				  GTK_SIGNAL_FUNC(schema_button_tree), 
+				  SCHEMATAB(tab)->treeroot);
+	       gtk_signal_connect(GTK_OBJECT(new_item), "button_release_event",
+				  GTK_SIGNAL_FUNC(schema_button_tree), 
+				  SCHEMATAB(tab)->treeroot);
 	       gtk_tree_append(GTK_TREE(server_subtree), new_item);
 	       gtk_widget_show(new_item);
 
 	       new_subtree = gtk_tree_new();
 	       GTK_WIDGET_UNSET_FLAGS(new_subtree, GTK_CAN_FOCUS);
-	       gtk_object_set_data(GTK_OBJECT(new_subtree), "server", server);
+	       gtk_object_set_data_full(GTK_OBJECT(new_subtree), "server", 
+					server,
+					(GtkDestroyNotify) ldapserver_unref);
+	       ldapserver_ref(server);
+
 	       gtk_widget_show(new_subtree);
 	       gtk_tree_item_set_subtree(GTK_TREE_ITEM(new_item), new_subtree);
-	       gtk_signal_connect(GTK_OBJECT(new_subtree), "select_child",
-				  GTK_SIGNAL_FUNC(fill_at_detail_rightpane), hash);
+	       gtk_signal_connect(GTK_OBJECT(new_subtree), "select-child",
+				  GTK_SIGNAL_FUNC(fill_at_detail_rightpane), tab);
 
 	       tmp = ss->at;
 	       while(tmp) {
 		    at = (LDAPAttributeType *) tmp->data;
-		    if(at->at_names && at->at_names[0])
-			 for(i = 0; at->at_names[i]; i++)
+		    if(at->at_names && at->at_names[0]) {
+			 for(i = 0; at->at_names[i]; i++) {
 			      attach_schema_item(new_subtree, "at", at, at->at_names[i]);
-		    else
+			 }
+		    } else {
 			 attach_schema_item(new_subtree, "at", at, at->at_oid);
+		    }
 		    tmp = tmp->next;
 	       }
 	  }
@@ -221,26 +316,36 @@ void attach_server_schema(GtkWidget *item, GHashTable *hash)
 	       new_item = gtk_tree_item_new_with_label("matchingRules");
 	       GTK_WIDGET_UNSET_FLAGS(new_item, GTK_CAN_FOCUS);
 	       gtk_signal_connect(GTK_OBJECT(new_item), "button_press_event",
-				  GTK_SIGNAL_FUNC(schema_button_tree), NULL);
+				  GTK_SIGNAL_FUNC(schema_button_tree),
+				  SCHEMATAB(tab)->treeroot);
+	       gtk_signal_connect(GTK_OBJECT(new_item), "button_release_event",
+				  GTK_SIGNAL_FUNC(schema_button_tree),
+				  SCHEMATAB(tab)->treeroot);
 	       gtk_tree_append(GTK_TREE(server_subtree), new_item);
 	       gtk_widget_show(new_item);
 
 	       new_subtree = gtk_tree_new();
 	       GTK_WIDGET_UNSET_FLAGS(new_subtree, GTK_CAN_FOCUS);
-	       gtk_object_set_data(GTK_OBJECT(new_subtree), "server", server);
+	       gtk_object_set_data_full(GTK_OBJECT(new_subtree), "server", 
+					server,
+					(GtkDestroyNotify) ldapserver_unref);
+	       ldapserver_ref(server);
+
 	       gtk_widget_show(new_subtree);
 	       gtk_tree_item_set_subtree(GTK_TREE_ITEM(new_item), new_subtree);
-	       gtk_signal_connect(GTK_OBJECT(new_subtree), "select_child",
-				  GTK_SIGNAL_FUNC(fill_mr_detail_rightpane), hash);
+	       gtk_signal_connect(GTK_OBJECT(new_subtree), "select-child",
+				  GTK_SIGNAL_FUNC(fill_mr_detail_rightpane), tab);
 
 	       tmp = ss->mr;
 	       while(tmp) {
 		    mr = (LDAPMatchingRule *) tmp->data;
-		    if(mr->mr_names && mr->mr_names[0])
-			 for(i = 0; mr->mr_names[i]; i++)
+		    if(mr->mr_names && mr->mr_names[0]) {
+			 for(i = 0; mr->mr_names[i]; i++) {
 			      attach_schema_item(new_subtree, "mr", mr, mr->mr_names[i]);
-		    else
+			 }
+		    } else {
 			 attach_schema_item(new_subtree, "mr", mr, mr->mr_oid);
+		    }
 		    tmp = tmp->next;
 	       }
 	  }
@@ -249,17 +354,25 @@ void attach_server_schema(GtkWidget *item, GHashTable *hash)
 	       new_item = gtk_tree_item_new_with_label("ldapSyntaxes");
 	       GTK_WIDGET_UNSET_FLAGS(new_item, GTK_CAN_FOCUS);
 	       gtk_signal_connect(GTK_OBJECT(new_item), "button_press_event",
-				  GTK_SIGNAL_FUNC(schema_button_tree), NULL);
+				  GTK_SIGNAL_FUNC(schema_button_tree),
+				  SCHEMATAB(tab)->treeroot);
+	       gtk_signal_connect(GTK_OBJECT(new_item), "button_release_event",
+				  GTK_SIGNAL_FUNC(schema_button_tree),
+				  SCHEMATAB(tab)->treeroot);
 	       gtk_tree_append(GTK_TREE(server_subtree), new_item);
 	       gtk_widget_show(new_item);
 
 	       new_subtree = gtk_tree_new();
 	       GTK_WIDGET_UNSET_FLAGS(new_subtree, GTK_CAN_FOCUS);
-	       gtk_object_set_data(GTK_OBJECT(new_subtree), "server", server);
+	       gtk_object_set_data_full(GTK_OBJECT(new_subtree), "server",
+					server,
+					(GtkDestroyNotify) ldapserver_unref);
+	       ldapserver_ref(server);
+
 	       gtk_widget_show(new_subtree);
 	       gtk_tree_item_set_subtree(GTK_TREE_ITEM(new_item), new_subtree);
-	       gtk_signal_connect(GTK_OBJECT(new_subtree), "select_child",
-				  GTK_SIGNAL_FUNC(fill_s_detail_rightpane), hash);
+	       gtk_signal_connect(GTK_OBJECT(new_subtree), "select-child",
+				  GTK_SIGNAL_FUNC(fill_s_detail_rightpane), tab);
 
 	       tmp = ss->s;
 	       while(tmp) {
@@ -273,18 +386,24 @@ void attach_server_schema(GtkWidget *item, GHashTable *hash)
 
      set_normalcursor();
 
+     error_flush(attach_context);
 }
 
 
 /* more of a macro really, this makes inserting duplicates easier */
-void attach_schema_item(GtkWidget *tree, char *type, void *schemaobject, char *itemname)
+static void attach_schema_item(GtkWidget *tree, const char *type, 
+			       void *schemaobject, char *itemname)
 {
      GtkWidget *new_item;
 
      new_item = gtk_tree_item_new_with_label(itemname);
      GTK_WIDGET_UNSET_FLAGS(new_item, GTK_CAN_FOCUS);
      gtk_signal_connect(GTK_OBJECT(new_item), "button_press_event",
-			GTK_SIGNAL_FUNC(schema_button_tree), NULL);
+			GTK_SIGNAL_FUNC(schema_button_tree),
+			tree);
+     gtk_signal_connect(GTK_OBJECT(new_item), "button_release_event",
+			GTK_SIGNAL_FUNC(schema_button_tree),
+			tree);
      gtk_object_set_data(GTK_OBJECT(new_item), type, schemaobject);
      gtk_widget_show(new_item);
      /* need gtk_tree_insert_sorted() here :-( */
@@ -293,7 +412,8 @@ void attach_schema_item(GtkWidget *tree, char *type, void *schemaobject, char *i
 }
 
 
-gboolean schema_button_tree(GtkWidget *tree_item, GdkEventButton *event)
+static gboolean schema_button_tree(GtkWidget *tree_item, GdkEventButton *event,
+				   GtkWidget *root)
 {
      GtkWidget *root_menu, *menu, *menu_item;
 
@@ -305,10 +425,12 @@ gboolean schema_button_tree(GtkWidget *tree_item, GdkEventButton *event)
 	  menu = gtk_menu_new();
 	  gtk_menu_item_set_submenu(GTK_MENU_ITEM(root_menu), menu);
 
+#if GTK_MAJOR < 2
 	  menu_item = gtk_tearoff_menu_item_new();
 	  gtk_menu_append(GTK_MENU(menu), menu_item);
 	  gtk_widget_set_sensitive(menu_item, FALSE);
 	  gtk_widget_show(menu_item);
+#endif
 
 	  /* Refresh */
 	  menu_item = gtk_menu_item_new_with_label(_("Refresh"));
@@ -334,22 +456,28 @@ gboolean schema_button_tree(GtkWidget *tree_item, GdkEventButton *event)
 			 event->button, event->time);
 
 	  return(TRUE);
-     }
-     else if(event->type == GDK_BUTTON_PRESS && event->button == 2) {
+     } else if(event->type == GDK_BUTTON_RELEASE && event->button == 2) {
 	  popup_detail_callback(NULL, tree_item);
 	  return(TRUE);
+#if GTK_MAJOR >= 2
+	  /* hack, hack, hack */
+     } else if(event->type == GDK_BUTTON_RELEASE && event->button == 1) {
+	  gtk_tree_select_child(GTK_TREE(root), tree_item);
+	  return(TRUE);
+#endif
      }
 
      return(FALSE);
 }
 
 
-void schema_refresh_server(GtkWidget *widget, GtkWidget *item)
+static void schema_refresh_server(GtkWidget *widget, GtkWidget *item)
 {
      GList *sel;
      GtkTree *tree;
      GtkWidget *new_subtree;
      struct ldapserver *server;
+     struct tab *tab;
 
      if( (server = gtk_object_get_data(GTK_OBJECT(item), "server")) == NULL)
 	  return;
@@ -360,7 +488,11 @@ void schema_refresh_server(GtkWidget *widget, GtkWidget *item)
      if(tree) {
 
 	  /* this is a workaround -- lots of GTK warnings if I don't do this :-( */
+#if GTK_MAJOR < 2
 	  sel = GTK_TREE_SELECTION(tree);
+#else
+	  sel = GTK_TREE_SELECTION_OLD(tree);
+#endif
 	  while(sel) {
 	       if(sel->data)
 		    gtk_tree_unselect_child(GTK_TREE(tree), sel->data);
@@ -371,14 +503,14 @@ void schema_refresh_server(GtkWidget *widget, GtkWidget *item)
 
 	  new_subtree = gtk_tree_new();
 	  gtk_tree_item_set_subtree(GTK_TREE_ITEM(item), new_subtree);
+	  tab = gtk_object_get_data(GTK_OBJECT(item), "tab");
 	  gtk_signal_connect(GTK_OBJECT(item), "expand",
-			     GTK_SIGNAL_FUNC(attach_server_schema), NULL);
+			     GTK_SIGNAL_FUNC(attach_server_schema), tab);
      }
-
 }
 
 
-void make_detail_notebook(GHashTable *hash)
+static void make_detail_notebook(struct tab *tab)
 {
      GtkWidget *rightpane_notebook, *rightpane_vbox;
      GtkWidget *oc_vbox, *at_vbox, *mr_vbox, *s_vbox;
@@ -392,17 +524,17 @@ void make_detail_notebook(GHashTable *hash)
      gtk_widget_show(at_vbox);
      gtk_widget_show(mr_vbox);
      gtk_widget_show(s_vbox);
-     g_hash_table_insert(hash, "oc_vbox", oc_vbox);
-     g_hash_table_insert(hash, "at_vbox", at_vbox);
-     g_hash_table_insert(hash, "mr_vbox", mr_vbox);
-     g_hash_table_insert(hash, "s_vbox", s_vbox);
+     SCHEMATAB(tab)->oc_vbox = oc_vbox;
+     SCHEMATAB(tab)->at_vbox = at_vbox;
+     SCHEMATAB(tab)->mr_vbox = mr_vbox;
+     SCHEMATAB(tab)->s_vbox = s_vbox;
 
-     rightpane_vbox = g_hash_table_lookup(hash, "rightpane_vbox");
+     rightpane_vbox = SCHEMATAB(tab)->rightpane_vbox;
      rightpane_notebook = gtk_notebook_new();
      gtk_box_pack_start(GTK_BOX(rightpane_vbox), rightpane_notebook, TRUE, TRUE, 0);
      GTK_WIDGET_UNSET_FLAGS(GTK_NOTEBOOK(rightpane_notebook), GTK_CAN_FOCUS);
      gtk_widget_show(rightpane_notebook);
-     g_hash_table_insert(hash, "rightpane_notebook", rightpane_notebook);
+     SCHEMATAB(tab)->rightpane_notebook = rightpane_notebook;
 
      /* Objectclasses tab */
      dummy_vbox = gtk_vbox_new(FALSE, 0);
@@ -443,7 +575,7 @@ void make_detail_notebook(GHashTable *hash)
 }
 
 
-void popup_detail_callback(GtkWidget *dummy, GtkWidget *widget)
+static void popup_detail_callback(GtkWidget *dummy, GtkWidget *widget)
 {
      LDAPObjectClass *oc;
      LDAPAttributeType *at;
@@ -476,9 +608,11 @@ void popup_detail_callback(GtkWidget *dummy, GtkWidget *widget)
 }
 
 
-void popup_detail(enum schema_detail_type type, struct ldapserver *server, void *detail)
+void popup_detail(enum schema_detail_type type,
+		  struct ldapserver *server, void *detail)
 {
      GtkWidget *window, *vbox;
+     int error_context;
 
      window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
      gtk_window_set_title(GTK_WINDOW(window), _("GQ"));
@@ -489,6 +623,9 @@ void popup_detail(enum schema_detail_type type, struct ldapserver *server, void
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(window), vbox);
 
+     error_context = error_new_context(_("Showing schema details"),
+				       window);
+
      switch(type) {
      case SCHEMA_TYPE_OC:
 	  gtk_window_set_title(GTK_WINDOW(window), _("GQ: objectclass"));
@@ -498,7 +635,7 @@ void popup_detail(enum schema_detail_type type, struct ldapserver *server, void
      case SCHEMA_TYPE_AT:
 	  gtk_window_set_title(GTK_WINDOW(window), _("GQ: attribute type"));
 	  make_at_detail(vbox);
-	  fill_at_detail(vbox, server, detail);
+	  fill_at_detail(error_context, vbox, server, detail);
 	  break;
      case SCHEMA_TYPE_MR:
 	  gtk_window_set_title(GTK_WINDOW(window), _("GQ: matching rule"));
@@ -514,10 +651,11 @@ void popup_detail(enum schema_detail_type type, struct ldapserver *server, void
 
      gtk_widget_show(window);
 
+     error_flush(error_context);
 }
 
 
-void make_oc_detail(GtkWidget *target_oc_vbox)
+static void make_oc_detail(GtkWidget *target_oc_vbox)
 {
      GtkWidget *label, *entry, *combo, *check, *clist;
      GtkWidget *hbox1, *hbox2, *vbox1, *table1, *scrwin, *paned;
@@ -640,8 +778,11 @@ void make_oc_detail(GtkWidget *target_oc_vbox)
      /* right half */
 
      paned = gtk_vpaned_new();
+#if GTK_MAJOR < 1
      gtk_paned_set_gutter_size(GTK_PANED(paned), 0);
      gtk_paned_set_handle_size(GTK_PANED(paned), 0);
+#endif
+
      gtk_widget_show(paned);
      gtk_box_pack_start(GTK_BOX(hbox1), paned, TRUE, TRUE, 10);
 
@@ -649,7 +790,8 @@ void make_oc_detail(GtkWidget *target_oc_vbox)
      scrwin = gtk_scrolled_window_new(NULL, NULL);
      gtk_widget_show(scrwin);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrwin),
-                                    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
+				    GTK_POLICY_AUTOMATIC, 
+				    GTK_POLICY_AUTOMATIC);
 
      /* just about right for four lines in the top clist */
      gtk_widget_set_usize(scrwin, -1, 93);
@@ -687,7 +829,7 @@ void make_oc_detail(GtkWidget *target_oc_vbox)
 }
 
 
-void fill_oc_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item, GHashTable *hash)
+static void fill_oc_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item, struct tab *tab)
 {
      GtkWidget *rightpane_notebook, *oc_vbox;
      LDAPObjectClass *oc;
@@ -699,33 +841,35 @@ void fill_oc_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item, GHashTa
      if( (oc = gtk_object_get_data(GTK_OBJECT(tree_item), "oc")) == NULL)
 	  return;
 
-     if(g_hash_table_lookup(hash, "cur_oc") == oc)
+     if(SCHEMATAB(tab)->cur_oc == oc)
 	  return;
 
-     g_hash_table_insert(hash, "cur_oc", oc);
+     SCHEMATAB(tab)->cur_oc = oc;
 
-     oc_vbox = g_hash_table_lookup(hash, "oc_vbox");
+     oc_vbox = SCHEMATAB(tab)->oc_vbox;
      if(oc_vbox == NULL) {
-	  make_detail_notebook(hash);
-	  oc_vbox = g_hash_table_lookup(hash, "oc_vbox");
+	  make_detail_notebook(tab);
+	  oc_vbox = SCHEMATAB(tab)->oc_vbox;
      }
 
      fill_oc_detail(oc_vbox, server, oc);
 
      /* switch to objectClasses tab */
-     rightpane_notebook = g_hash_table_lookup(hash, "rightpane_notebook");
+     rightpane_notebook = SCHEMATAB(tab)->rightpane_notebook;
      gtk_notebook_set_page(GTK_NOTEBOOK(rightpane_notebook), 0);
 
 }
 
 
-void fill_oc_detail(GtkWidget *target_oc_vbox, struct ldapserver *server,
-		    LDAPObjectClass *oc)
+static void fill_oc_detail(GtkWidget *target_oc_vbox,
+			   struct ldapserver *server,
+			   LDAPObjectClass *oc)
 {
      GList *list;
      GtkWidget *entry, *combo, *check, *clist;
      int i;
-     char *kind, *dummy[2];
+     const char *kind;
+     char *dummy[2];
 
      /* left half */
 
@@ -779,7 +923,10 @@ void fill_oc_detail(GtkWidget *target_oc_vbox, struct ldapserver *server,
 
      /* Required attributes */
      clist = (GtkWidget *) gtk_object_get_data(GTK_OBJECT(target_oc_vbox), "required");
-     gtk_object_set_data(GTK_OBJECT(clist), "server", server);
+     gtk_object_set_data_full(GTK_OBJECT(clist), "server", 
+			      server,(GtkDestroyNotify)  ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_clist_freeze(GTK_CLIST(clist));
      gtk_clist_clear(GTK_CLIST(clist));
      i = 0;
@@ -793,7 +940,10 @@ void fill_oc_detail(GtkWidget *target_oc_vbox, struct ldapserver *server,
 
      /* Allowed attributes */
      clist = (GtkWidget *) gtk_object_get_data(GTK_OBJECT(target_oc_vbox), "allowed");
-     gtk_object_set_data(GTK_OBJECT(clist), "server", server);
+     gtk_object_set_data_full(GTK_OBJECT(clist), "server", 
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_clist_freeze(GTK_CLIST(clist));
      gtk_clist_clear(GTK_CLIST(clist));
      i = 0;
@@ -807,7 +957,7 @@ void fill_oc_detail(GtkWidget *target_oc_vbox, struct ldapserver *server,
 }
 
 
-void make_at_detail(GtkWidget *target_vbox)
+static void make_at_detail(GtkWidget *target_vbox)
 {
      GtkWidget *hbox1, *hbox2, *vbox1, *vbox2, *table1, *table2, *scrwin;
      GtkWidget *label, *entry, *combo, *check, *clist;
@@ -1067,11 +1217,13 @@ void make_at_detail(GtkWidget *target_vbox)
 }
 
 
-void fill_at_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item, GHashTable *hash)
+static void fill_at_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item,
+				     struct tab *tab)
 {
      GtkWidget *rightpane_notebook, *at_vbox;
      LDAPAttributeType *at;
      struct ldapserver *server;
+     int error_context; 
 
      if( (server = gtk_object_get_data(GTK_OBJECT(treeroot), "server")) == NULL)
 	  return;
@@ -1079,34 +1231,39 @@ void fill_at_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item, GHashTa
      if( (at = gtk_object_get_data(GTK_OBJECT(tree_item), "at")) == NULL)
 	  return;
 
-     if(g_hash_table_lookup(hash, "cur_at") == at)
+     if(SCHEMATAB(tab)->cur_at == at)
 	  return;
 
-     g_hash_table_insert(hash, "cur_at", at);
+     SCHEMATAB(tab)->cur_at = at;
 
-     at_vbox = g_hash_table_lookup(hash, "at_vbox");
+     at_vbox = SCHEMATAB(tab)->at_vbox;
      if(at_vbox == NULL) {
-	  make_detail_notebook(hash);
-	  at_vbox = g_hash_table_lookup(hash, "at_vbox");
+	  make_detail_notebook(tab);
+	  at_vbox = SCHEMATAB(tab)->at_vbox;
      }
 
-     fill_at_detail(at_vbox, server, at);
+     error_context = error_new_context(_("Attribute details"), at_vbox);
+
+     fill_at_detail(error_context, at_vbox, server, at);
 
      /* switch to Attribute types tab */
-     rightpane_notebook = g_hash_table_lookup(hash, "rightpane_notebook");
+     rightpane_notebook = SCHEMATAB(tab)->rightpane_notebook;
      gtk_notebook_set_page(GTK_NOTEBOOK(rightpane_notebook), 1);
 
+     error_flush(error_context);
+
 }
 
 
-void fill_at_detail(GtkWidget *target_vbox, struct ldapserver *server,
-		    LDAPAttributeType *at)
+static void fill_at_detail(int error_context, GtkWidget *target_vbox,
+			   struct ldapserver *server, LDAPAttributeType *at)
 {
      GList *list;
      GString *syntax;
      GtkWidget *combo, *entry, *check, *clist;
      int i;
-     char *usage, *dummy[2];
+     const char *usage;
+     char *dummy[2];
 
      /* left half */
 
@@ -1189,12 +1346,15 @@ void fill_at_detail(GtkWidget *target_vbox, struct ldapserver *server,
      gtk_entry_set_text(GTK_ENTRY(entry), syntax->str);
 
      clist = (GtkWidget *) gtk_object_get_data(GTK_OBJECT(target_vbox), "usedoc");
-     gtk_object_set_data(GTK_OBJECT(clist), "server", server);
+     gtk_object_set_data_full(GTK_OBJECT(clist), "server",
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_clist_freeze(GTK_CLIST(clist));
      gtk_clist_clear(GTK_CLIST(clist));
 
      if(at->at_names && at->at_names[0]) {
-	  list = find_oc_by_at(server, at->at_names[0]);
+	  list = find_oc_by_at(error_context, server, at->at_names[0]);
 	  if(list) {
 	       i = 0;
 	       dummy[1] = NULL;
@@ -1212,7 +1372,7 @@ void fill_at_detail(GtkWidget *target_vbox, struct ldapserver *server,
 }
 
 
-void make_mr_detail(GtkWidget *target_vbox)
+static void make_mr_detail(GtkWidget *target_vbox)
 {
      GtkWidget *hbox1, *vbox1, *table1, *hbox2, *vbox2, *scrwin;
      GtkWidget *label, *combo, *entry, *check, *clist;
@@ -1340,7 +1500,8 @@ void make_mr_detail(GtkWidget *target_vbox)
 }
 
 
-void fill_mr_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item, GHashTable *hash)
+static void fill_mr_detail_rightpane(GtkWidget *treeroot, 
+				     GtkWidget *tree_item, struct tab *tab)
 {
      GtkWidget *rightpane_notebook, *mr_vbox;
      LDAPMatchingRule *mr;
@@ -1352,28 +1513,28 @@ void fill_mr_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item, GHashTa
      if( (mr = gtk_object_get_data(GTK_OBJECT(tree_item), "mr")) == NULL)
 	  return;
 
-     if(g_hash_table_lookup(hash, "cur_mr") == mr)
+     if(SCHEMATAB(tab)->cur_mr == mr)
 	  return;
 
-     g_hash_table_insert(hash, "cur_mr", mr);
+     SCHEMATAB(tab)->cur_mr = mr;
 
-     mr_vbox = g_hash_table_lookup(hash, "mr_vbox");
+     mr_vbox = SCHEMATAB(tab)->mr_vbox;
      if(mr_vbox == NULL) {
-	  make_detail_notebook(hash);
-	  mr_vbox = g_hash_table_lookup(hash, "mr_vbox");
+	  make_detail_notebook(tab);
+	  mr_vbox = SCHEMATAB(tab)->mr_vbox;
      }
 
      fill_mr_detail(mr_vbox, server, mr);
 
      /* switch to Matching rules tab */
-     rightpane_notebook = g_hash_table_lookup(hash, "rightpane_notebook");
+     rightpane_notebook = SCHEMATAB(tab)->rightpane_notebook;
      gtk_notebook_set_page(GTK_NOTEBOOK(rightpane_notebook), 2);
 
 }
 
 
-void fill_mr_detail(GtkWidget *target_vbox, struct ldapserver *server,
-		    LDAPMatchingRule *mr)
+static void fill_mr_detail(GtkWidget *target_vbox, struct ldapserver *server,
+			   LDAPMatchingRule *mr)
 {
      GList *list;
      GtkWidget *combo, *entry, *check, *clist;
@@ -1409,7 +1570,10 @@ void fill_mr_detail(GtkWidget *target_vbox, struct ldapserver *server,
      /* right half */
 
      clist = (GtkWidget *) gtk_object_get_data(GTK_OBJECT(target_vbox), "usedin");
-     gtk_object_set_data(GTK_OBJECT(clist), "server", server);
+     gtk_object_set_data_full(GTK_OBJECT(clist), "server", 
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_clist_freeze(GTK_CLIST(clist));
      gtk_clist_clear(GTK_CLIST(clist));
 
@@ -1431,7 +1595,7 @@ void fill_mr_detail(GtkWidget *target_vbox, struct ldapserver *server,
 }
 
 
-void make_s_detail(GtkWidget *target_vbox)
+static void make_s_detail(GtkWidget *target_vbox)
 {
      GtkWidget *hbox1, *vbox1, *table1, *scrwin, *paned;
      GtkWidget *label, *entry, *clist;
@@ -1490,8 +1654,11 @@ void make_s_detail(GtkWidget *target_vbox)
      /* right half */
 
      paned = gtk_vpaned_new();
+#if GTK_MAJOR < 1
      gtk_paned_set_gutter_size(GTK_PANED(paned), 0);
      gtk_paned_set_handle_size(GTK_PANED(paned), 0);
+#endif
+
      gtk_paned_set_position(GTK_PANED(paned), 157);
      gtk_widget_show(paned);
      gtk_box_pack_start(GTK_BOX(hbox1), paned, TRUE, TRUE, 10);
@@ -1537,7 +1704,8 @@ void make_s_detail(GtkWidget *target_vbox)
 }
 
 
-void fill_s_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item, GHashTable *hash)
+static void fill_s_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item,
+				    struct tab *tab)
 {
      GtkWidget *rightpane_notebook, *s_vbox;
      LDAPSyntax *s;
@@ -1549,28 +1717,28 @@ void fill_s_detail_rightpane(GtkWidget *treeroot, GtkWidget *tree_item, GHashTab
      if( (s = gtk_object_get_data(GTK_OBJECT(tree_item), "s")) == NULL)
 	  return;
 
-     if(g_hash_table_lookup(hash, "cur_s") == s)
+     if(SCHEMATAB(tab)->cur_s == s)
 	  return;
 
-     g_hash_table_insert(hash, "cur_s", s);
+     SCHEMATAB(tab)->cur_s = s;
 
-     s_vbox = g_hash_table_lookup(hash, "s_vbox");
+     s_vbox = SCHEMATAB(tab)->s_vbox;
      if(s_vbox == NULL) {
-	  make_detail_notebook(hash);
-	  s_vbox = g_hash_table_lookup(hash, "s_vbox");
+	  make_detail_notebook(tab);
+	  s_vbox = SCHEMATAB(tab)->s_vbox;
      }
 
      fill_s_detail(s_vbox, server, s);
 
      /* switch to Syntaxes tab */
-     rightpane_notebook = g_hash_table_lookup(hash, "rightpane_notebook");
+     rightpane_notebook = SCHEMATAB(tab)->rightpane_notebook;
      gtk_notebook_set_page(GTK_NOTEBOOK(rightpane_notebook), 3);
 
 }
 
 
-void fill_s_detail(GtkWidget *target_vbox, struct ldapserver *server,
-		   LDAPSyntax *s)
+static void fill_s_detail(GtkWidget *target_vbox, struct ldapserver *server,
+			  LDAPSyntax *s)
 {
      GList *list;
      GtkWidget *entry, *clist;
@@ -1587,7 +1755,10 @@ void fill_s_detail(GtkWidget *target_vbox, struct ldapserver *server,
 
      /* Used in attribute types */
      clist = (GtkWidget *) gtk_object_get_data(GTK_OBJECT(target_vbox), "usedat");
-     gtk_object_set_data(GTK_OBJECT(clist), "server", server);
+     gtk_object_set_data_full(GTK_OBJECT(clist), "server", 
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_clist_freeze(GTK_CLIST(clist));
      gtk_clist_clear(GTK_CLIST(clist));
 
@@ -1607,7 +1778,10 @@ void fill_s_detail(GtkWidget *target_vbox, struct ldapserver *server,
 
      /* Used in matching rules */
      clist = (GtkWidget *) gtk_object_get_data(GTK_OBJECT(target_vbox), "usedmr");
-     gtk_object_set_data(GTK_OBJECT(clist), "server", server);
+     gtk_object_set_data_full(GTK_OBJECT(clist), "server", 
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_clist_freeze(GTK_CLIST(clist));
      gtk_clist_clear(GTK_CLIST(clist));
 
@@ -1679,21 +1853,27 @@ void select_at_from_clist(GtkWidget *clist, gint row, gint column,
      /* double click or single middle button click */
      if( (event->type == GDK_BUTTON_RELEASE && event->button == 2) ||
 	 (event->type == GDK_2BUTTON_PRESS && event->button == 1)) {
+	  gint i;
 
-	  if( (server = gtk_object_get_data(GTK_OBJECT(clist), "server")) == NULL)
+	  if( (server = gtk_object_get_data(GTK_OBJECT(clist), "server")) == NULL) {
 	       return;
+	  }
 
-	  if(server->ss == NULL || server->ss->at == NULL)
+	  if(server->ss == NULL || server->ss->at == NULL) {
 	       return;
+	  }
 
 	  gtk_clist_get_text(GTK_CLIST(clist), row, column, &attrname);
-	  list = server->ss->at;
-	  while(list) {
+	  for(list = server->ss->at; list; list = list->next) {
 	       at = list->data;
-	       if(at && at->at_names && at->at_names[0] &&
-		  !strcasecmp(at->at_names[0], attrname))
-		    break;
-	       list = list->next;
+	       //g_return_val_if_fail(at && at->at_names); // FIXME: we should require this
+	       for(i=0; at && at->at_names && at->at_names[i]; i++) {
+	            if(!strcasecmp(at->at_names[i], attrname))
+		         break; // the for loop
+	       }
+	       if(at && at->at_names && at->at_names[i]) { // at_names[i] == NULL after the for loop
+		       break; // the while loop
+	       }
 	  }
 
 	  if(list && at)
@@ -1743,11 +1923,80 @@ void select_mr_from_clist(GtkWidget *clist, gint row, gint column,
 
 #endif /* HAVE_LDAP_STR2OBJECTCLASS */
 
-void cleanup_schema_mode(void)
+static void destroy_schema_mode(GtkWidget *w, struct tab *tab)
 {
+     assert(tab);
+     /* the whole thing is (almost) stateless! */
 
-     /* the whole thing is stateless! */
+     g_free(tab->modeinfo);
+     tab->modeinfo = NULL;
+     
+     g_free(tab);
+}
+
+
+void update_schema_serverlist(struct tab *tab)
+{
+#if 0
+     GtkTree *treeroot = SCHEMATAB(tab)->treeroot;
+
+     struct ldapserver *server;
+/*       cb_server_list csl; */
+/*       GList *l; */
+
+
+
+     /* walk the list of ldapservers, add any not yet in the show list */
+     /* NOTE: This should take care of renaming a server as well, somehow */
+     for (server = config.ldapservers ; server ; server = server->next) {
+	  if (server_node_from_server(ctree, server) == NULL)
+	       add_single_schema_server(ctree, server);
+     }
+
+
+
+     csl.list = NULL;
+     gtk_ctree_pre_recursive_to_depth(ctree,
+				      NULL, /* root node */
+				      1,    /* max depth */
+				      (GQTreeWidgetFunc) get_current_servers_list_check_func,
+				      &csl);
+
+     /* walk list of shown servers, remove any no longer in config */
+     for (l = csl.list ; l ; l=l->next) {
+	  int found = 0;
+	  struct ldapserver *thisserver = l->data;
+	  GQTreeWidgetNode *node = server_node_from_server(ctree, thisserver);
+
+	  /* is this server still in the list of configured servers? */
+	  struct ldapserver *server;
+	  for (server = config.ldapservers ; server ; server = server->next) {
+	       if (thisserver == server) {
+		    found = 1;
+		    break;
+	       }
+	  }
+
+	  /* deleted servers */
+	  if (!found) {
+	       if (node) gq_tree_remove_node(ctree, node);
+	  } else {
+	       /* renamed servers ? */
+	       char *currtext = gq_tree_get_node_text (ctree, node);
+	       if (currtext && strcmp(currtext, thisserver->name) != 0) {
+		    gq_tree_set_node_text (ctree, node, thisserver->name);
+	       }
+	  }
+     }
+     if (csl.list) g_list_free(csl.list);
+
+#endif
 
 }
 
 
+/* 
+   Local Variables:
+   c-basic-offset: 5
+   End:
+*/
