@@ -1,6 +1,7 @@
 /*
     GQ -- a GTK-based LDAP client
-    Copyright (C) 1998-2002 Bert Vermeulen
+    Copyright (C) 1998-2003 Bert Vermeulen
+    Copyright (C) 2002-2003 Peter Stamfest
 
     This program is released under the Gnu General Public License with
     the additional exemption that compiling, linking, and/or using
@@ -21,6 +22,7 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+/* $Id: template.c,v 1.17 2006/04/05 00:53:55 herzi Exp $ */
 
 #include <config.h>
 
@@ -45,16 +47,42 @@
 #include "tdefault.h"
 #include "errorchain.h"
 #include "debug.h"
+#include "i18n.h"
+#include "state.h"
 
 
-extern struct gq_config config;
 extern GtkWidget *current_template_clist;
+static GtkWidget *templatewin;
 
+static void build_used_attrs_list(GtkWidget *window, int context);
 
-GtkWidget *templatewin;
+/* malloc and init gq_template struct */
+struct gq_template *new_template(void)
+{
+     struct gq_template *tmpl;
 
+     tmpl = g_malloc0(sizeof(struct gq_template));
 
-void create_template_edit_window(struct ldapserver *server, char *templatename)
+     tmpl->name = g_strdup("");
+     tmpl->objectclasses = NULL;
+
+     return tmpl;
+}
+
+void free_template(struct gq_template *tmpl)
+{
+     g_free_if(tmpl->name);
+     
+     if (tmpl->objectclasses) {
+	  g_list_foreach(tmpl->objectclasses, (GFunc) g_free, NULL);
+     }
+     g_list_free(tmpl->objectclasses);
+     g_free(tmpl);
+}
+
+void create_template_edit_window(struct ldapserver *server, 
+				 const char *templatename,
+				 GtkWidget *modalFor)
 {
      GList *list;
      GString *gmessage;
@@ -63,39 +91,56 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
      GtkWidget *oclist, *templatelist, *clist, *button, *arrow, *label, *entry;
      struct server_schema *ss;
      struct gq_template *tmpl;
-     char *otitle[1] = { "Available objectclasses" };
-     char *ttitle[1] = { "Template objectclasses" };
-     char *rtitle[1] = { "Required attributes" };
-     char *atitle[1] = { "Allowed attributes" };
+     char *otitle[1] = { _("Available objectclasses") };
+     char *ttitle[1] = { _("Template objectclasses") };
+     char *rtitle[1] = { _("Required attributes") };
+     char *atitle[1] = { _("Allowed attributes") };
+
+     int create_context = error_new_context(_("Opening template window"),
+					    modalFor);
 
      gmessage = NULL;
-     ss = get_schema(server);
+     ss = get_schema(create_context, server);
      if(ss == NULL) {
 	  gmessage = g_string_sized_new(64);
-	  g_string_sprintf(gmessage, "no schema information found for server %s"
-			   , server->name);
+	  g_string_sprintf(gmessage,
+			   _("No schema information found for server %s"),
+			   server->name);
      }
      else if(ss->oc == NULL) {
 	  gmessage = g_string_sized_new(64);
-	  g_string_sprintf(gmessage, "no objectclass information found for server %s",
+	  g_string_sprintf(gmessage,
+			   _("No objectclass information found for server %s"),
 			   server->name);
      }
      else if(ss->at == NULL) {
 	  gmessage = g_string_sized_new(64);
-	  g_string_sprintf(gmessage, "no attribute type information found for server %s",
+	  g_string_sprintf(gmessage,
+			   _("No attribute type information found for server %s"),
 			   server->name);
      }
 
      if(gmessage) {
-	  error_popup("Error getting schema", gmessage->str);
+	  error_push(create_context, gmessage->str);
 	  g_string_free(gmessage, TRUE);
-	  return;
+
+	  goto done;
      }
 
-     templatewin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
-     gtk_object_set_data(GTK_OBJECT(templatewin), "server", server);
+     templatewin = stateful_gtk_window_new(GTK_WINDOW_TOPLEVEL,
+					   "template", 700, 450); 
+     if (modalFor) {
+	  assert(GTK_IS_WINDOW(modalFor));
+	  gtk_window_set_modal(GTK_WINDOW(templatewin), TRUE);
+	  gtk_window_set_transient_for(GTK_WINDOW(templatewin),
+				       GTK_WINDOW(modalFor));
+     }
+
+     gtk_object_set_data_full(GTK_OBJECT(templatewin), "server", 
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_container_border_width(GTK_CONTAINER(templatewin), 10);
-     gtk_window_set_default_size(GTK_WINDOW(templatewin), 700, 450);
      gtk_signal_connect(GTK_OBJECT(templatewin), "key_press_event",
 			(GtkSignalFunc) close_on_esc, (gpointer) templatewin);
 
@@ -120,8 +165,13 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
 
      oclist = gtk_clist_new_with_titles(1, otitle);
      GTK_CLIST(oclist)->button_actions[1] = GTK_BUTTON_SELECTS;
+#ifdef OLD_FOCUS_HANDLING
      GTK_WIDGET_UNSET_FLAGS(oclist, GTK_CAN_FOCUS);
-     gtk_object_set_data(GTK_OBJECT(oclist), "server", server);
+#endif
+     gtk_object_set_data_full(GTK_OBJECT(oclist), "server", 
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_object_set_data(GTK_OBJECT(templatewin), "oclist", oclist);
      gtk_clist_set_compare_func(GTK_CLIST(oclist), clistcasecmp);
      gtk_clist_column_titles_passive(GTK_CLIST(oclist));
@@ -142,12 +192,14 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
      /* right arrow */
      button = gtk_button_new();
      gtk_widget_set_usize(button, 20, 30);
+#ifdef OLD_FOCUS_HANDLING
      GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
+#endif
      gtk_object_set_data(GTK_OBJECT(button), "direction", "right");
      arrow = gtk_arrow_new(GTK_ARROW_RIGHT, GTK_SHADOW_OUT);
      gtk_widget_show(arrow);
      gtk_container_add(GTK_CONTAINER(button), arrow);
-     gtk_signal_connect(GTK_OBJECT(button), "pressed",
+     gtk_signal_connect(GTK_OBJECT(button), "clicked",
 			(GtkSignalFunc) arrow_button_callback, templatewin);
      gtk_widget_show(button);
      gtk_box_pack_start(GTK_BOX(vbox3), button, TRUE, TRUE, 10);
@@ -155,12 +207,14 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
      /* left arrow */
      button = gtk_button_new();
      gtk_widget_set_usize(button, 20, 30);
+#ifdef OLD_FOCUS_HANDLING
      GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
+#endif
      gtk_object_set_data(GTK_OBJECT(button), "direction", "left");
      arrow = gtk_arrow_new(GTK_ARROW_LEFT, GTK_SHADOW_OUT);
      gtk_widget_show(arrow);
      gtk_container_add(GTK_CONTAINER(button), arrow);
-     gtk_signal_connect(GTK_OBJECT(button), "pressed",
+     gtk_signal_connect(GTK_OBJECT(button), "clicked",
 			(GtkSignalFunc) arrow_button_callback, templatewin);
      gtk_widget_show(button);
      gtk_box_pack_start(GTK_BOX(vbox3), button, TRUE, TRUE, 10);
@@ -174,8 +228,13 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
 
      templatelist = gtk_clist_new_with_titles(1, ttitle);
      GTK_CLIST(templatelist)->button_actions[1] = GTK_BUTTON_SELECTS;
+#ifdef OLD_FOCUS_HANDLING
      GTK_WIDGET_UNSET_FLAGS(templatelist, GTK_CAN_FOCUS);
-     gtk_object_set_data(GTK_OBJECT(templatelist), "server", server);
+#endif
+     gtk_object_set_data_full(GTK_OBJECT(templatelist), "server", 
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_object_set_data(GTK_OBJECT(templatewin), "templatelist", templatelist);
      gtk_clist_column_titles_passive(GTK_CLIST(templatelist));
      gtk_signal_connect(GTK_OBJECT(templatelist), "select_row",
@@ -195,8 +254,13 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
      gtk_widget_show(scrwin);
      clist = gtk_clist_new_with_titles(1, rtitle);
      GTK_CLIST(clist)->button_actions[1] = GTK_BUTTON_SELECTS;
+#ifdef OLD_FOCUS_HANDLING
      GTK_WIDGET_UNSET_FLAGS(clist, GTK_CAN_FOCUS);
-     gtk_object_set_data(GTK_OBJECT(clist), "server", server);
+#endif
+     gtk_object_set_data_full(GTK_OBJECT(clist), "server",
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_object_set_data(GTK_OBJECT(templatewin), "reqdattr", clist);
      gtk_clist_column_titles_passive(GTK_CLIST(clist));
      gtk_signal_connect(GTK_OBJECT(clist), "select_row",
@@ -212,8 +276,13 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
      gtk_widget_show(scrwin);
      clist = gtk_clist_new_with_titles(1, atitle);
      GTK_CLIST(clist)->button_actions[1] = GTK_BUTTON_SELECTS;
+#ifdef OLD_FOCUS_HANDLING
      GTK_WIDGET_UNSET_FLAGS(clist, GTK_CAN_FOCUS);
-     gtk_object_set_data(GTK_OBJECT(clist), "server", server);
+#endif
+     gtk_object_set_data_full(GTK_OBJECT(clist), "server", 
+			      server, (GtkDestroyNotify) ldapserver_unref);
+     ldapserver_ref(server);
+
      gtk_object_set_data(GTK_OBJECT(templatewin), "allowedattr", clist);
      gtk_clist_column_titles_passive(GTK_CLIST(clist));
      gtk_signal_connect(GTK_OBJECT(clist), "select_row",
@@ -230,8 +299,10 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
      /* Input defaults */
      button = gtk_button_new_with_label("  Edit input defaults  ");
      GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_DEFAULT);
+#ifdef OLD_FOCUS_HANDLING
      GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
-     gtk_signal_connect(GTK_OBJECT(button), "pressed",
+#endif
+     gtk_signal_connect(GTK_OBJECT(button), "clicked",
 			(GtkSignalFunc) create_tdefault_edit_window,
 			(gpointer) templatewin);
      gtk_widget_show(button);
@@ -243,7 +314,7 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
      gtk_box_pack_start(GTK_BOX(vbox1), hbox3, FALSE, FALSE, 10);
 
      /* Template name */
-     label = gtk_label_new("Template name");
+     label = gtk_label_new(_("Template name"));
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(hbox3), label, FALSE, FALSE, 0);
 
@@ -256,14 +327,18 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
      gtk_widget_show(entry);
      gtk_box_pack_start(GTK_BOX(hbox3), entry, FALSE, FALSE, 10);
 
-     hbox4 = gtk_hbox_new(FALSE, 0);
+     hbox4 = gtk_hbutton_box_new();
      gtk_widget_show(hbox4);
      gtk_box_pack_start(GTK_BOX(vbox1), hbox4, FALSE, FALSE, 10);
 
      /* Save */
-     button = gtk_button_new_with_label("  Save  ");
+#if GTK_MAJOR >= 2
+     button = gtk_button_new_from_stock(GTK_STOCK_SAVE);
+#else
+     button = gq_button_new_with_label(_("_Save"));
+#endif
      GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
-     gtk_signal_connect(GTK_OBJECT(button), "pressed",
+     gtk_signal_connect(GTK_OBJECT(button), "clicked",
 			(GtkSignalFunc) save_template_callback,
 			(gpointer) templatewin);
      gtk_widget_show(button);
@@ -271,33 +346,38 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
      gtk_widget_grab_default(button);
 
      /* Cancel */
-     button = gtk_button_new_with_label("  Cancel  ");
-     GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
-     gtk_signal_connect_object(GTK_OBJECT(button), "pressed",
+#if GTK_MAJOR >= 2
+     button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
+#else
+     button = gq_button_new_with_label(_("_Cancel"));
+#endif
+     gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
 			       (GtkSignalFunc) gtk_widget_destroy,
 			       (gpointer) templatewin);
      gtk_widget_show(button);
      gtk_box_pack_end(GTK_BOX(hbox4), button, FALSE, FALSE, 0);
-     gtk_signal_connect_object(GTK_OBJECT(templatewin), "destroy",
-			       (GtkSignalFunc) delete_edit_window,
-			       (gpointer) templatewin);
 
      fill_new_template(templatewin);
 
      /* default title, in case editing doesn't work out */
-     gtk_window_set_title(GTK_WINDOW(templatewin), "GQ: create new template");
+     gtk_window_set_title(GTK_WINDOW(templatewin),
+			  _("GQ: create new template"));
 
      if(templatename) {
-	  gtk_window_set_title(GTK_WINDOW(templatewin), "GQ: edit template");
-	  gtk_object_set_data(GTK_OBJECT(templatewin), "templatename", g_strdup(templatename));
+	  gtk_window_set_title(GTK_WINDOW(templatewin),
+			       _("GQ: edit template"));
+	  gtk_object_set_data_full(GTK_OBJECT(templatewin),
+				   "templatename", 
+				   g_strdup(templatename), g_free);
 	  tmpl = find_template_by_name(templatename);
 	  if(tmpl) {
 	       gtk_entry_set_text(GTK_ENTRY(entry), tmpl->name);
 	       list = tmpl->objectclasses;
 	       while(list) {
 		    /* a bit slow maybe */
-		    move_objectclass(oclist, templatelist, (char *) list->data);
-		    build_used_attrs_list(templatewin);
+		    move_objectclass(oclist, templatelist, 
+				     (const char *) list->data);
+		    build_used_attrs_list(templatewin, create_context);
 		    list = list->next;
 	       }
 	  }
@@ -306,28 +386,15 @@ void create_template_edit_window(struct ldapserver *server, char *templatename)
 	  /* a new template has the "top" objectclass selected
 	     by default (LDAP V3 requirement) */
 	  move_objectclass(oclist, templatelist, "top");
-	  build_used_attrs_list(templatewin);
+	  build_used_attrs_list(templatewin, create_context);
      }
 
      gtk_widget_show(templatewin);
 
      gtk_widget_grab_focus(entry);
 
-}
-
-
-/*
- * templatename holds the name of the existing template the editor is using.
- * this is g_strdup'ed so need to free it before the window goes
- */
-gboolean delete_edit_window(GtkWidget *window)
-{
-     char *templatename;
-
-     if( (templatename = gtk_object_get_data(GTK_OBJECT(window), "templatename")))
-	  g_free(templatename);
-
-     return(FALSE);
+ done:
+     error_flush(create_context);
 }
 
 
@@ -377,6 +444,7 @@ void arrow_button_callback(GtkWidget *widget, GtkWidget *window)
 {
      GtkWidget *oclist, *templatelist;
      char *direction, *objectclass;
+     int context;
 
      if( (oclist = gtk_object_get_data(GTK_OBJECT(window), "oclist")) == NULL)
 	  return;
@@ -397,8 +465,13 @@ void arrow_button_callback(GtkWidget *widget, GtkWidget *window)
 	  gtk_clist_sort(GTK_CLIST(oclist));
      }
 
-     build_used_attrs_list(window);
+     context = 
+	  error_new_context(_("Adding/deleting objectclass to/from template"),
+			    window);
 
+     build_used_attrs_list(window, context);
+
+     error_flush(context);
 }
 
 
@@ -421,7 +494,8 @@ int clistcasecmp(GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2)
 /*
  * move item by name from source clist to destination clist
  */
-void move_objectclass(GtkWidget *source, GtkWidget *destination, char *objectclass)
+void move_objectclass(GtkWidget *source, GtkWidget *destination, 
+		      const char *objectclass)
 {
      int row;
      char *dummy[1];
@@ -446,7 +520,7 @@ void move_objectclass(GtkWidget *source, GtkWidget *destination, char *objectcla
  * build "Required attributes" and "Allowed attributes" clists from
  * current objectclasses in "Template" clist
  */
-void build_used_attrs_list(GtkWidget *window)
+static void build_used_attrs_list(GtkWidget *window, int context)
 {
      LDAPObjectClass *oc;
      GtkWidget *templatelist, *reqdlist, *allowedlist;
@@ -467,7 +541,7 @@ void build_used_attrs_list(GtkWidget *window)
      if( (allowedlist = gtk_object_get_data(GTK_OBJECT(window), "allowedattr")) == NULL)
 	  return;
 
-     ss = get_server_schema(server);
+     ss = get_server_schema(context, server);
 
      gtk_clist_freeze(GTK_CLIST(reqdlist));
      gtk_clist_clear(GTK_CLIST(reqdlist));
@@ -509,69 +583,87 @@ void build_used_attrs_list(GtkWidget *window)
  */
 void save_template_callback(GtkWidget *widget, GtkWidget *window)
 {
-     GList *list;
      struct gq_template *old_tmpl, *tmpl;
      char *old_template_name;
+     int error_context;
+     gboolean save_ok = FALSE;
+
+     error_context = error_new_context(_("Saving template"), widget);
 
      tmpl = window2template(window);
      if(tmpl == NULL) {
 	  /* window2template() failed */
-	  error_popup("Weirdness!", "Template was not saved!");
-	  return;
+	  error_push(error_context, _("Could not create new template."));
+	  goto done;
      }
 
      if(strlen(tmpl->name) == 0) {
 	  /* need a name for this template, warning popup */
-	  list = g_list_append(NULL, "You need to enter a name for the new template.");
-	  list = g_list_append(list, "Please enter one and try again.");
-	  warning_popup(list);
-	  return;
+	  error_push(error_context,
+		     _("You need to enter a name for the new template."));
+	  free_template(tmpl);
+	  goto done;
      }
 
      if( (old_template_name = gtk_object_get_data(GTK_OBJECT(window), "templatename")) == NULL) {
 	  /* add new template to list, but make sure the template name is unique */
 	  if( (find_template_by_name(tmpl->name)) == NULL) {
-	       config.templates = g_list_append(config.templates, tmpl);
+	       config->templates = g_list_append(config->templates, tmpl);
+	  } else {
+	       error_push(error_context,
+			  _("A template with that name already exists!"));
+	       free_template(tmpl);
+	       goto done;
 	  }
-	  else {
-	       list = g_list_append(NULL, "A template by that name already exists!");
-	       list = g_list_append(list, "Please choose another name, or delete "
-				    "the existing template");
-	       warning_popup(list);
-	       return;
+	  save_ok = save_config_ext(error_context);
+	  if (!save_ok) {
+	       /* save failed - undo changes */
+	       config->templates = g_list_remove(config->templates, tmpl);
+	       free_template(tmpl);
 	  }
-     }
-     else {
-
+     } else {
+	  GList *l;
 	  /* update existing template */
 	  old_tmpl = find_template_by_name(old_template_name);
+	  
+	  /* In order to encapsulate knowledge about the internal
+	     structure of a template object, updating is done via
+	     delete/insert. This is less optimal than the old
+	     behaviour to change the filter in-place
+
+	     (especially in case some other part of gq holds a pointer
+	     to the template)
+	  */
 	  if(old_tmpl) {
+	       l = g_list_find(config->templates, old_tmpl);
+	       l->data = tmpl;
 
-	       g_list_free(old_tmpl->objectclasses);
-	       old_tmpl->objectclasses = tmpl->objectclasses;
-
-	       if(strcmp(old_template_name, tmpl->name)) {
-		    /* template name was changed too */
-		    strncpy(old_tmpl->name, tmpl->name, 128);
+	       save_ok = save_config_ext(error_context);
+	       if (save_ok) {
+		    free_template(old_tmpl);
+	       } else {
+		    l->data = old_tmpl;
+		    free_template(tmpl);
+	       }
+	  } else {
+	       /* hmmmm... template was deleted while editing it. Just
+		  add it in then. */
+	       config->templates = g_list_append(config->templates, tmpl);
+	       save_ok = save_config_ext(error_context);
+	       if (!save_ok) {
+		    config->templates = g_list_remove(config->templates, tmpl);
+		    free_template(tmpl);
 	       }
-
-	       /* tmpl is not inserted into global config, so we don't need it any more */
-	       FREE(tmpl, "struct gq_template");
-
 	  }
-	  else {
-	       /* hmmmm... template was deleted while editing it. Just add it in then */
-	       config.templates = g_list_append(config.templates, tmpl);
-	  }
-
      }
 
-     save_config();
-
-     fill_clist_templates(current_template_clist);
-
-     gtk_widget_destroy(window);
+     if (save_ok) {
+	  fill_clist_templates(current_template_clist);
+	  gtk_widget_destroy(window);
+     }
 
+ done:
+     error_flush(error_context);
 }
 
 
@@ -586,6 +678,7 @@ struct gq_template *window2template(GtkWidget *window)
      struct gq_template *tmpl;
      int i;
      char *objectclass;
+     const char *c;
 
      if( (templatelist = gtk_object_get_data(GTK_OBJECT(window), "templatelist")) == NULL)
 	  return(NULL);
@@ -593,10 +686,12 @@ struct gq_template *window2template(GtkWidget *window)
      if( (templatenamebox = gtk_object_get_data(GTK_OBJECT(window), "templatenamebox")) == NULL)
 	  return(NULL);
 
-     if( (tmpl = MALLOC(sizeof(struct gq_template), "struct gq_template")) == NULL)
+     if( (tmpl = new_template()) == NULL)
 	  return(NULL);
 
-     strncpy(tmpl->name, gtk_entry_get_text(GTK_ENTRY(templatenamebox)), 128);
+     
+     c = gtk_entry_get_text(GTK_ENTRY(templatenamebox));
+     g_free_and_dup(tmpl->name, c);
 
      list = NULL;
      for(i = 0; i < GTK_CLIST(templatelist)->rows; i++) {
@@ -622,7 +717,7 @@ void fill_clist_templates(GtkWidget *clist)
      gtk_clist_freeze(GTK_CLIST(clist));
      gtk_clist_clear(GTK_CLIST(clist));
 
-     list = config.templates;
+     list = config->templates;
      while(list) {
 	  if( (tmpl = (struct gq_template *) list->data)) {
 	       dummy[0] = tmpl->name;
@@ -658,7 +753,7 @@ char *get_clist_selection(GtkWidget *clist)
  * find selected item in clist by row string
  * return row number
  */
-int get_clist_row_by_text(GtkWidget *clist, char *text)
+int get_clist_row_by_text(GtkWidget *clist, const char *text)
 {
      int i, row;
      char *str;
@@ -677,3 +772,9 @@ int get_clist_row_by_text(GtkWidget *clist, char *text)
 
 
 #endif /* HAVE_LDAP_STR2OBJECTCLASS */
+
+/* 
+   Local Variables:
+   c-basic-offset: 5
+   End:
+ */
