@@ -1,6 +1,7 @@
 /*
     GQ -- a GTK-based LDAP client
-    Copyright (C) 1998-2001 Bert Vermeulen
+    Copyright (C) 1998-2003 Bert Vermeulen
+    Copyright (C) 2002-2003 Peter Stamfest
 
     This program is released under the Gnu General Public License with
     the additional exemption that compiling, linking, and/or using
@@ -21,8 +22,29 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+/* $Id: tinput.h,v 1.15 2003/11/02 07:20:44 stamfest Exp $ */
 
-GList *formfill_from_template(struct ldapserver *schemaserver, struct gq_template *template);
-GList *formfill_from_entry_objectclass(struct ldapserver *server, char *dn);
-GList *add_attrs_by_oc(struct ldapserver *server, GList *oclist);
-GList *add_schema_attrs(struct ldapserver *server, GList *value_list);
+#ifndef GQ_TINPUT_H_INCLUDED
+#define GQ_TINPUT_H_INCLUDED
+
+#include <glib.h>
+#include "common.h"
+#include "template.h"
+
+GList *formfill_from_template(int error_context, 
+			      struct ldapserver *schemaserver, 
+			      struct gq_template *template);
+
+GList *formfill_from_entry_objectclass(int error_context,
+				       struct ldapserver *server,
+				       const char *dn);
+
+GList *add_attrs_by_oc(int error_context,
+		       struct ldapserver *server,
+		       GList *oclist);
+
+GList *add_schema_attrs(int error_context,
+			struct ldapserver *server,
+			GList *value_list);
+
+#endif /* GQ_TINPUT_H_INCLUDED */
